# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QComboBox

from ...helpers import BaseDialog

from ..helpers import tagged_items

from .helpers import delete_version, init_version_selector


class DeleteVersionDialog(BaseDialog):
    """ This class implements the dialog for deleting a version. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        self._version = QComboBox()
        layout.addWidget(self._version)

    def set_fields(self):
        """ Set the dialog's fields from the project. """

        init_version_selector(self._version, self.model)

    def get_fields(self):
        """ Update the project from the dialog's fields. """

        delete_version(self._version.currentText(), self.shell,
                migrate_items=True)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .annos import Annos
from .tagged import Tagged
from .workflow import Workflow


@dataclass
class Code(Annos, Tagged, Workflow):
    """ This class implements APIs that can be annotated, are subject to
    version control and a workflow.
    """



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QSplitter

from .control_widget import ControlWidget
from .sources_widget import SourcesWidget


class ScannerGui(QSplitter):
    """ This class is a widget that implements a scanner's GUI. """

    def __init__(self, tool):
        """ Initialise the widget. """

        super().__init__()

        self.sources_widget = SourcesWidget(tool)
        self.addWidget(self.sources_widget)

        self.control_widget = ControlWidget(tool)
        self.addWidget(self.control_widget)

    def restore_state(self, settings):
        """ Restore the widget's state. """

        settings.beginGroup('control')
        self.control_widget.restore_state(settings)
        settings.endGroup()

        settings.beginGroup('sources')
        self.sources_widget.restore_state(settings)
        settings.endGroup()

        state = settings.value('splitter')
        if state is not None:
            self.restoreState(state)

    def save_state(self, settings):
        """ Save the widget's state. """

        settings.setValue('splitter', self.saveState())

        settings.beginGroup('sources')
        self.sources_widget.save_state(settings)
        settings.endGroup()

        settings.beginGroup('control')
        self.control_widget.save_state(settings)
        settings.endGroup()



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..annos import Annos

from .adapt import adapt
from .base_adapter import AttributeType, BaseApiAdapter


class ArgumentAdapter(BaseApiAdapter):
    """ This is the Argument adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'default':      AttributeType.STRING,
        'name':         AttributeType.STRING,
        'pydefault':    AttributeType.STRING,
        'pytype':       AttributeType.STRING,
        'type':         AttributeType.STRING,
        'unnamed':      AttributeType.BOOL,
    }

    def __eq__(self, other):
        """ Compare for C/C++ equality. """

        arg = self.model
        other_arg = other.model

        if type(arg) is not type(other_arg):
            return False

        if self.expand_type(arg.type) != self.expand_type(other_arg.type):
            return False

        if arg.default != other_arg.default:
            return False

        return True

    def as_py_str(self):
        """ Return the Python representation of the argument. """

        arg = self.model

        s = self.expand_type(arg.pytype if arg.pytype != '' else arg.type,
                name=arg.name)

        s += adapt(arg, Annos).as_str()

        if arg.pydefault != '':
            s += ' = ' + arg.pydefault
        elif arg.default != '':
            s += ' = ' + arg.default

        return s

    def as_str(self):
        """ Return the standard string representation. """

        arg = self.model

        s = self.expand_type(arg.type, name=arg.name)

        if arg.default != '':
            s += ' = ' + arg.default

        return s

    def generate_sip(self, sip_file, output):
        """ Generate the .sip file content. """

        output.write(self.as_py_str())

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        super().load(element, project, ui)

        adapt(self.model, Annos).load(element, project, ui)

    def save(self, output):
        """ Save the model to an output file. """

        argument = self.model

        output.write('<Argument')
        adapt(argument, Annos).save_attributes(output)
        self.save_attribute('type', argument.type, output)
        self.save_bool('unnamed', output)
        self.save_str('name', output)
        self.save_str('default', output)
        self.save_str('pydefault', output)
        self.save_str('pytype', output)

        # Note that we are assuming Annos does not have any subelements.
        output.write('/>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QCheckBox, QDialog

from ....helpers import BaseDialog


class PlatformsDialog(BaseDialog):
    """ This class implements the dialog for selecting a number of platforms.
    """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        project = self.shell.project

        self._platforms = []

        for platform in project.platforms:
            check_box = QCheckBox(platform)
            layout.addWidget(check_box)
            self._platforms.append((check_box, platform))

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        for check_box, platform in self._platforms:
            check_box.setChecked(platform in self.model.platforms)

    def get_fields(self):
        """ Update the API item from the dialog's fields. """

        platforms = []

        for check_box, platform in self._platforms:
            if check_box.isChecked():
                platforms.append(platform)

        self.model.platforms = platforms

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from .header_directory import (get_platform_name, get_supported_platforms,
        header_directory_platform)
from .version_map import VersionMap
from .version_range import version_range



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .access import Access
from .callable import Callable


@dataclass
class OperatorMethod(Callable, Access):
    """ This class implements a C++ class operator. """

    # This is set if the operator is abstract.
    abstract: bool = False

    # This is set if the operator is const.
    const: bool = False

    # The optional %VirtualCatcherCode.
    virtcode: str = ''

    # This is set if the operator is virtual.
    virtual: bool = False



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from enum import auto, Enum

from PyQt6.QtCore import QSettings, Qt
from PyQt6.QtWidgets import (QDockWidget, QMainWindow, QMenuBar, QMessageBox,
        QWhatsThis)

from .._version import version

from .shell_tool import ShellTool, ShellToolLocation


class EventType(Enum):
    """ The different project-related event types.  These are only used to
    update the different parts of the GUI and so there are many similar changes
    that don't have an event because there is no consequent impact on the GUI.
    """

    # The status of an API model has changed.  The event argument is the API.
    API_STATUS = auto()

    # The versions of an API model has changed.  The event argument is the API.
    API_VERSIONS = auto()

    # An API model has been added to a container model.  The event argument is
    # a 2-tuple of the container and API.
    CONTAINER_API_ADD = auto()

    # An API model has been deleted from a container model.  The event argument
    # is a 2-tuple of the container and API.
    CONTAINER_API_DELETE = auto()

    # A feature has been added.  The event argument is the name of the feature.
    FEATURE_ADD = auto()

    # A feature has been deleted.  The event argument is the name of the
    # feature.
    FEATURE_DELETE = auto()

    # A feature has been renamed.  The event argument is a 2-tuple of the old
    # and new names.
    FEATURE_RENAME = auto()

    # A message has been logged.  The message is the event argument.
    LOG_MESSAGE = auto()

    # A module model has been added.  The event argument is the module.
    MODULE_ADD = auto()

    # A module model has been deleted.  The event argument is the module.
    MODULE_DELETE = auto()

    # A module model has been renamed.  The event argument is the module.
    MODULE_RENAME = auto()

    # A platform has been added.  The event argument is the name of the
    # platform.
    PLATFORM_ADD = auto()

    # A platform has been deleted.  The event argument is the name of the
    # platform.
    PLATFORM_DELETE = auto()

    # A platform has been renamed.  The event argument is a 2-tuple of the old
    # and new names.
    PLATFORM_RENAME = auto()

    # A new project has been loaded.  There is no event argument.
    PROJECT_NEW = auto()

    # A project has been renamed.  There is no event argument.
    PROJECT_RENAME = auto()

    # The project's root module has been renamed.  There is no event argument.
    PROJECT_ROOT_MODULE_RENAME = auto()

    # A version has been added.  The event argument is the name of the version.
    VERSION_ADD = auto()

    # A version has been deleted.  The event argument is the name of the
    # version.
    VERSION_DELETE = auto()

    # A version has been renamed.  The event argument is a 2-tuple of the old
    # and new names.
    VERSION_RENAME = auto()


class Shell:
    """ This class encapsulates a collection of tools. """

    # Map tool locations to Qt dock areas.
    _LOCATION_MAP = {
        ShellToolLocation.LEFT: Qt.DockWidgetArea.LeftDockWidgetArea,
        ShellToolLocation.RIGHT: Qt.DockWidgetArea.RightDockWidgetArea,
        ShellToolLocation.TOP: Qt.DockWidgetArea.TopDockWidgetArea,
        ShellToolLocation.BOTTOM: Qt.DockWidgetArea.BottomDockWidgetArea,
    }

    def __init__(self, *tool_factories):
        """ Initialise the shell. """

        self._project = None

        # Create the widget that implements the shell.
        self.widget = _ShellWidget(self._handle_close_event)

        # Create the tools.
        self._tools = []
        menus = {}
        view_actions = []

        for tool_factory in tool_factories:
            tool = tool_factory(self)

            if isinstance(tool, ShellTool):
                if tool.location is ShellToolLocation.CENTRE:
                    self.widget.setCentralWidget(tool.widget)
                else:
                    dock_widget = QDockWidget(tool.title, objectName=tool.name)
                    dock_widget.setWidget(tool.widget)
                    self.widget.addDockWidget(
                            self._LOCATION_MAP[tool.location], dock_widget)
                    view_actions.append(dock_widget.toggleViewAction())

            # Add any actions to the menus.
            menu_name, actions = tool.actions
            if menu_name:
                menu_actions = menus.setdefault(menu_name, [])
                if len(menu_actions) != 0:
                    # This will turn into a separator.
                    menu_actions.append(None)

                menu_actions.extend(actions)

            self._tools.append(tool)

        # Add any view actions to the View menu.
        if len(view_actions) != 0:
            all_view_actions = menus.setdefault("View", [])
            if len(all_view_actions) != 0:
                all_view_actions.append(None)

            all_view_actions.extend(view_actions)

        # Create the menu bar.
        menu_bar = QMenuBar()

        def add_menu(menu_name, actions):
            menu = menu_bar.addMenu(menu_name)

            for action in actions:
                if action is None:
                    menu.addSeparator()
                else:
                    menu.addAction(action)

        # Make sure any well known menus are created in the expected order.
        for menu_name in ("File", "Edit", "View", "Tools"):
            actions = menus.pop(menu_name, None)
            if actions is not None:
                add_menu(menu_name, actions)

        # Now do any remaining menus.
        for menu_name, actions in menus.items():
            add_menu(menu_name, actions)

        help_menu = menu_bar.addMenu("&Help")
        help_menu.addAction("About msip...", self._about)
        help_menu.addAction(QWhatsThis.createAction(help_menu))

        self.widget.setMenuBar(menu_bar)

        # Restore the state now the GUI has been built.
        settings = QSettings()

        for tool in self._tools:
            settings.beginGroup(tool.name)
            tool.restore_state(settings)
            settings.endGroup()

        settings.beginGroup('metasip.shell')

        setting = settings.value('state')
        if setting is not None:
            self.widget.restoreState(setting)

        setting = settings.value('geometry')
        if setting is not None:
            self.widget.restoreGeometry(setting)

        settings.endGroup()

    @property
    def dirty(self):
        """ Get the project's dirty state. """

        return self._project.dirty

    @dirty.setter
    def dirty(self, state):
        """ Set the dirty state. """

        # We do this here in case the project name has changed.
        self.widget.setWindowTitle(self._project.name + '[*]')

        self._project.dirty = state
        self.widget.setWindowModified(state)

    def handle_project_dialog(self, title, dialog_factory):
        """ Handle a dialog that will update some aspect of a project. """

        dialog = dialog_factory(self._project, title, self)

        if dialog.update():
            self.dirty = True

    def log(self, message):
        """ Log a message. """

        self.notify(EventType.LOG_MESSAGE, message)

    def notify(self, event_type, event_arg=None):
        """ Notify all tools about a project-specific event. """

        for tool in self._tools:
            tool.event(event_type, event_arg)

    @property
    def project(self):
        """ Get the current project. """

        return self._project

    @project.setter
    def project(self, project):
        """ Set the current project. """

        self._project = project
        self.notify(EventType.PROJECT_NEW)
        self.dirty = project.dirty

    def save_project(self):
        """ Return True if any current project was saved. """

        # Ask the user unless the project hasn't been modified.
        if not self._project.dirty:
            return True

        button = QMessageBox.question(self.widget, "Quit",
                "The project has been modified. Do you want to save or discard the changes?",
                QMessageBox.StandardButton.Cancel | QMessageBox.StandardButton.Discard | QMessageBox.StandardButton.Save)

        if button is QMessageBox.StandardButton.Cancel:
            return False

        if button is QMessageBox.StandardButton.Save:
            save_unsuccessful = False

            for tool in self._tools:
                if not tool.save_data():
                    save_unsuccessful = True

            if save_unsuccessful:
                return False

        return True

    def show(self):
        """ Make the shell visible. """

        self.widget.show()

    def _about(self):
        """ Tell the user about the application. """

        QMessageBox.about(self.widget, "About msip",
f"""This is msip v{version}, part of MetaSIP.

msip is a tool for creating .sip files from C/C++ header files.
""")

    def _handle_close_event(self):
        """ Handle a close event and return True if the event should be
        accepted.
        """

        if self.save_project():
            # Save the settings.
            settings = QSettings()

            settings.beginGroup('metasip.shell')
            settings.setValue('geometry', self.widget.saveGeometry())
            settings.setValue('state', self.widget.saveState())
            settings.endGroup()

            for tool in self._tools:
                settings.beginGroup(tool.name)
                tool.save_state(settings)
                settings.endGroup()

            return True

        return False


class _ShellWidget(QMainWindow):
    """ The widget that implements the shell. """

    def __init__(self, close_event_handler):
        """ Initialise the widget. """

        super().__init__()

        self._close_event_handler = close_event_handler

    def closeEvent(self, event):
        """ Reimplemented to save the state of the GUI. """

        if self._close_event_handler():
            event.accept()
        else:
            event.ignore()



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..sip_file import SipFile

from .adapt import adapt
from .base_adapter import AttributeType, BaseAdapter


class ModuleAdapter(BaseAdapter):
    """ This is the Module adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'directives':           AttributeType.LITERAL,
        'imports':              AttributeType.STRING_LIST,
        'name':                 AttributeType.STRING,
        'outputdirsuffix':      AttributeType.STRING,       # Removed in v0.17
        'pyssizetclean':        AttributeType.BOOL,
        'uselimitedapi':        AttributeType.BOOL,
        'virtualerrorhandler':  AttributeType.STRING,
    }

    def as_str(self):
        """ Return the standard string representation. """

        return self.model.name

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        module = self.model

        super().load(element, project, ui)

        callsuperinit = int(element.get('callsuperinit', '-1'))
        if callsuperinit < 0:
            callsuperinit = 'undefined'
        elif callsuperinit == 0:
            callsuperinit = 'no'
        else:
            callsuperinit = 'yes'

        module.callsuperinit = callsuperinit

        # Prior to v0.17 'keyword_arguments' was hardcoded as "Optional".
        module.keywordarguments = element.get('keywordarguments',
                'Optional' if project.version < (0, 17) else '')

        for subelement in element:
            if subelement.tag == 'SipFile':
                sip_file = SipFile()
                adapt(sip_file).load(subelement, project, ui)
                module.content.append(sip_file)

    def save(self, output):
        """ Save the model to an output file. """

        module = self.model

        output.write(f'<Module name="{module.name}"')

        if module.callsuperinit != 'undefined':
            self.save_attribute('callsuperinit',
                    '1' if module.callsuperinit == 'yes' else '0', output)

        self.save_str('keywordarguments', output)
        self.save_str('virtualerrorhandler', output)
        self.save_bool('uselimitedapi', output)
        self.save_bool('pyssizetclean', output)
        self.save_str_list('imports', output)
        output.write('>\n')
        output += 1

        self.save_literal('directives', output)

        for sip_file in module.content:
            adapt(sip_file).save(output)

        output -= 1
        output.write('</Module>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


import sys


def get_platform_name():
    """ Return the name of the current platform. """

    if sys.platform == 'darwin':
        return 'macOS'

    if sys.platform == 'win32':
        return 'Windows'

    platform_name = 'Linux'


def get_supported_platforms():
    """ Return the names of supported host platforms. """

    return ('Linux', 'macOS', 'Windows')


def header_directory_platform(header_directory, platform_name=None):
    """ Return a header directory's  Platform instance for a platform. """

    if platform_name is None:
        platform_name = get_platform_name()

    for platform in header_directory.platforms:
        if platform.name == platform_name:
            return platform



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from .base_adapter import AttributeType, BaseAdapter


class AccessAdapter(BaseAdapter):
    """ This is the Access adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'access':   AttributeType.STRING,
    }

    def save_attributes(self, output):
        """ Save the XML attributes. """

        self.save_str('access', output)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..models import VersionRange


class VersionMap:
    """ This class is a tool for consolidating and normalising ranges of
    versions.
    """

    def __init__(self, project, version_ranges=None):
        """ Return a version map with each entry set to an initial value. """

        self._versions = project.versions
        self._initialise_map(False)

        if version_ranges is not None:
            self.update_from_version_ranges(version_ranges)

    def __bool__(self):
        """ Return True if the version map is unconditionally True. """

        return all(self._map)

    def __getitem__(self, version):
        """ Return the map value for a version. """

        return self._map[self._versions.index(version)]

    def __setitem__(self, version, value):
        """ Set the map value for a version. """

        self._map[self._versions.index(version)] = value

    def update_from_version_ranges(self, version_ranges):
        """ Update the version map from a list of version ranges. """

        # No version ranges means all versions are valid.
        if len(version_ranges) == 0:
            self._initialise_map(True)
            return

        for version_range in version_ranges:
            if version_range.startversion == '':
                start_idx = 0
            else:
                start_idx = self._versions.index(version_range.startversion)

            if version_range.endversion == '':
                end_idx = len(self._versions)
            else:
                end_idx = self._versions.index(version_range.endversion)

            for i in range(start_idx, end_idx):
                self._map[i] = True

    def as_version_ranges(self):
        """ Convert a version map to a list of version ranges.  An empty list
        means it is unconditionally True, None means it is unconditionally
        False.
        """

        # See if the item is valid for all versions.
        if all(self._map):
            return []

        # See if the item is valid for no versions.
        for v in self._map:
            if v:
                break
        else:
            return None

        # Construct the new list of version ranges.
        version_ranges = []
        vrange = None

        for idx, v in enumerate(self._map):
            if v:
                # Start a new version range if there isn't one currently.
                if vrange is None:
                    vrange = VersionRange()

                    if idx != 0:
                        vrange.startversion = self._versions[idx]
            elif vrange is not None:
                vrange.endversion = self._versions[idx]
                version_ranges.append(vrange)
                vrange = None

        if vrange is not None:
            version_ranges.append(vrange)

        return version_ranges

    def _initialise_map(self, initial_state):
        """ Initialise the map. """

        self._map = [initial_state for v in self._versions]



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass


@dataclass
class ExtendedAccess:
    """ This class is a mixin for APIs subject to the extended (i.e. Qt
    specific) C++ access specifiers.
    """

    # The access specifier.  Values are '' (meaning public), 'protected',
    # 'protected slots', 'private', 'public slots', 'signals'.
    access: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..code import Code
from ..docstring import Docstring
from ..extended_access import ExtendedAccess

from .adapt import adapt
from .base_adapter import AttributeType, BaseApiAdapter


class ManualCodeAdapter(BaseApiAdapter):
    """ This is the ManualCode adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'body':     AttributeType.LITERAL,
        'methcode': AttributeType.LITERAL,
        'precis':   AttributeType.STRING,
    }

    def as_str(self):
        """ Return the standard string representation. """

        return self.model.precis

    def generate_sip(self, sip_file, output):
        """ Generate the .sip file content. """

        manual_code = self.model

        nr_ends = self.version_start(output)

        if manual_code.body != '':
            output.write(f'// {manual_code.precis}\n{manual_code.body}\n',
                    indent=False)
        elif manual_code.precis.startswith('%'):
            output.write(manual_code.precis + '\n', indent=False)
        else:
            output.write(manual_code.precis + ';\n')

        adapt(manual_code, Docstring).generate_sip_directives(output)
        output.write_code_directive('%MethodCode', manual_code.methcode)

        self.version_end(nr_ends, output)

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        super().load(element, project, ui)

        adapt(self.model, Code).load(element, project, ui)
        adapt(self.model, Docstring).load(element, project, ui)
        adapt(self.model, ExtendedAccess).load(element, project, ui)

    def save(self, output):
        """ Save the model to an output file. """

        manual_code = self.model

        output.write('<ManualCode')
        adapt(manual_code, Code).save_attributes(output)
        adapt(manual_code, Docstring).save_attributes(output)
        adapt(manual_code, ExtendedAccess).save_attributes(output)
        self.save_attribute('precis', manual_code.precis, output)
        output.write('>\n')

        output += 1
        self.save_literal('body', output)
        adapt(manual_code, Code).save_subelements(output)
        adapt(manual_code, Docstring).save_subelements(output)
        adapt(manual_code, ExtendedAccess).save_subelements(output)
        self.save_literal('methcode', output)
        output -= 1

        output.write('</ManualCode>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


class BaseType:
    """ This class implements helpers for the BaseType annotation. """

    BASE_TYPES = ('Enum', 'Flag', 'IntEnum', 'UIntEnum', 'IntFlag')

    def __init__(self, combo):
        """ Initialise the combo-box. """

        combo.addItems(self.BASE_TYPES)

        self._combo = combo

    def setAnnotation(self, anno):
        """ Set the current annotation. """

        for index, base_type in enumerate(self.BASE_TYPES):
            if base_type == anno:
                self._combo.setCurrentIndex(index)
                break;

    def annotation(self, annos_list):
        """ Get the current annotation. """

        base_type = str(self._combo.currentText())

        if base_type != 'Enum':
            annos_list.append('BaseType=' + base_type)


class Encoding:
    """ This class implements helpers for the Encoding annotation. """

    ENCODINGS = ('Default', 'None', 'ASCII', 'Latin-1', 'UTF-8')

    def __init__(self, combo):
        """ Initialise the combo-box. """

        combo.addItems(self.ENCODINGS)

        self._combo = combo

    def setAnnotation(self, anno):
        """ Set the current annotation. """

        for index, encoding in enumerate(self.ENCODINGS):
            if encoding == anno:
                self._combo.setCurrentIndex(index)
                break

    def annotation(self, annos_list):
        """ Get the current annotation. """

        encoding = str(self._combo.currentText())

        if encoding != 'Default':
            annos_list.append(f'Encoding="{encoding}"')


def split_annos(anno_s):
    """ Split a number of annotations represented as a string into a sequence
    of 2-tuple name/value pairs.
    """

    # Handle the trvial case.
    if anno_s == '':
        return []

    # This makes parsing easier.
    anno_s += ','

    NAME, VALUE, END_QUOTE, COMMA = range(4)

    annos = []
    expecting = NAME
    part_start = 0
    name = None

    for i, ch in enumerate(anno_s):
        if expecting == END_QUOTE:
            if ch == '"':
                value = anno_s[part_start:i]
                annos.append((name.strip(), value))
                expecting = COMMA
        else:
            if ch == '=':
                name = anno_s[part_start:i]
                expecting = VALUE
                part_start = i + 1
            elif ch == '"':
                expecting = END_QUOTE
                part_start = i + 1
            elif ch == ',':
                if expecting == NAME:
                    name = anno_s[part_start:i].strip()
                    value = None
                elif expecting == VALUE:
                    value = anno_s[part_start:i].strip()

                annos.append((name, value))

                expecting = NAME
                part_start = i + 1

    return annos



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass


@dataclass
class Workflow:
    """ This class is a mixin for APIs that are subject to a workflow. """

    # The multiline comments included in generated .sip files.
    comments: str = ''

    # The workflow status of the API item.  Values are '', 'ignored',
    # 'removed', 'todo', and 'unknown'.
    status: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass
from typing import Callable, Optional

from PyQt6.QtCore import QByteArray, QMimeData, Qt
from PyQt6.QtGui import QDrag
from PyQt6.QtWidgets import (QApplication, QMenu, QMessageBox, QProgressDialog,
        QTreeWidget, QTreeWidgetItem, QTreeWidgetItemIterator)

from ....models import (Class, Constructor, Destructor, Method, Function, Enum,
        EnumValue, OperatorFunction, Access, OperatorMethod, ManualCode,
        Module, OpaqueClass, OperatorCast, Namespace, Tagged, Typedef,
        Variable)
from ....models.adapters import adapt

from ...helpers import warning

from .dialogs import (ArgumentPropertiesDialog, CallablePropertiesDialog,
        ClassPropertiesDialog, EnumPropertiesDialog,
        EnumMemberPropertiesDialog, FeaturesDialog, ManualCodeDialog,
        ModulePropertiesDialog, MoveHeaderDialog, NamespacePropertiesDialog,
        OpaqueClassPropertiesDialog, PlatformsDialog, ProjectPropertiesDialog,
        TypedefPropertiesDialog, VariablePropertiesDialog, VersionsDialog)

from .external_editor import ExternalEditor


class ApiEditor(QTreeWidget):
    """ This class implements the API editor. """

    MIME_FORMAT = 'application/x-item'

    # The column numbers.
    (NAME, ACCESS, STATUS, VERSIONS) = range(4)

    def __init__(self, shell):
        """ Initialise the editor. """

        super().__init__()

        self._shell = shell
        self._project_view = None

        # Tweak the tree widget.
        self.setHeaderLabels(("Name", "Access", "Status", "Versions"))
        self.setSelectionMode(self.SelectionMode.ExtendedSelection)
        self.setDragEnabled(True)
        self.viewport().setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setDragDropMode(self.DragDropMode.InternalMove)

        self.dragged = None

    def api_status(self, api):
        """ Handle the change of status of an API. """

        for view in self._all_views():
            if view.api is api:
                view.draw_status()
                break

    def api_versions(self, api):
        """ Handle the change of versions of an API. """

        for view in self._all_views():
            if view.api is api:
                view.draw_versions()
                break

    def container_api_add(self, container, api):
        """ Handle the addition of an API to a container. """

        for view in self._all_views():
            if view.api is container:
                view.api_add(api)
                break

    def container_api_delete(self, container, api):
        """ Handle the deletion of an API from a container. """

        for view in self._all_views():
            if view.api is container:
                view.api_delete(api)
                break

    def module_add(self, module):
        """ Handle the addition of a module. """

        self._project_view.module_add(module)

    def module_delete(self, module):
        """ Handle the deletion of a module. """

        self._project_view.module_delete(module)

    def module_rename(self, module):
        """ A module has been renamed. """

        self._project_view.module_rename(module)

    def restore_state(self, settings):
        """ Restore the widget's state. """

        state = settings.value('header')
        if state is not None:
            self.header().restoreState(state)

    def set_project(self):
        """ Set the current project. """

        self.clear()
        self._project_view = ProjectView(self._shell, self)

    def root_module_updated(self):
        """ The name of the root module has been updated. """

        self._project_view.root_module_updated()

    def save_state(self, settings):
        """ Save the widget's state. """

        settings.setValue('header', self.header().saveState())

    def contextMenuEvent(self, ev):
        """ Reimplemented to handle a context menu event. """

        # Find the view.
        view = self.itemAt(ev.pos())

        if view:
            # Get the list of siblings that are also selected.
            siblings = []

            parent = view.parent()
            if parent is not None:
                for sibling in parent.all_child_views():
                    if sibling is not view and sibling.isSelected() and not sibling.isHidden():
                        siblings.append(sibling)

            # Check it has a menu.
            options = view.get_menu(siblings)

            if options:
                # Create the menu.
                menu = QMenu()

                for option in options:
                    if option is None:
                        menu.addSeparator()
                        continue

                    action = menu.addAction(option.text, option.handler)

                    action.setEnabled(option.enabled)

                    if option.checked is not None:
                        action.setCheckable(True)
                        action.setChecked(option.checked)

                menu.exec(ev.globalPos())

                ev.accept()
                return

        super().contextMenuEvent(ev)

    def startDrag(self, actions):
        """ Reimplemented to start the drag of a view. """

        # Find the selected view.
        dragging = None

        it = QTreeWidgetItemIterator(self, QTreeWidgetItemIterator.Selected)
        itm = it.value()

        while itm is not None:
            # We can only drag one view.
            if dragging is not None:
                return

            dragging = itm

            it += 1
            itm = it.value()

        # Make sure there was something to drag.
        if dragging is None:
            return

        # Serialize the id() of the view being dragged.
        data = QByteArray.number(dragging.type())
        mime = QMimeData()
        mime.setData(self.MIME_FORMAT, data)

        drag = QDrag(self)
        drag.setMimeData(mime)
        drag.exec()

    def dragEnterEvent(self, ev):
        """ Reimplemented to validate a drag enter. """

        if ev.source() is self:
            ev.accept()
        else:
            ev.ignore()

    def dragMoveEvent(self, ev):
        """ Reimplemented to validate a drag move. """

        if ev.source() is self:
            ev.accept()
        else:
            ev.ignore()

    def dropEvent(self, ev):
        """ Reimplemented to handle a drop. """

        source_target = self._source_target(ev)

        if source_target is None:
            ev.ignore()
        else:
            source, target = source_target
            target.drop(source)
            ev.accept()

    def _source_target(self, ev):
        """ Return a 2-tuple of source and target views or None if the drop
        wasn't appropriate.
        """

        # Get the target.
        target = self.itemAt(ev.pos())

        # Check that the payload is a view.
        mime = ev.mimeData()
        if not mime.hasFormat(self.MIME_FORMAT):
            return None

        # Get the view from the payload.
        source_id = int(mime.data(self.MIME_FORMAT))
        source = None

        it = QTreeWidgetItemIterator(self, QTreeWidgetItemIterator.DragEnabled)
        itm = it.value()

        while itm is not None:
            if itm.type() == source_id:
                source = itm
                break

            it += 1
            itm = it.value()

        if source is None or source is target:
            return None

        # Ask the target if it can handle the source.
        if not (isinstance(target, DropSite) and target.droppable(source)):
            return None

        return source, target

    def _all_views(self, container_view=None):
        """ A generator for all API views. """

        if container_view is None:
            container_view = self._project_view

        for view in container_view.all_child_views():
            yield view

            for sub_view in self._all_views(view):
                yield sub_view


@dataclass
class MenuOption:
    """ This class specifies a menu option. """

    # The option text.
    text: str

    # The option handler.
    handler: Callable

    # Set if the option is checked.
    checked: Optional[bool] = None

    # Set if the option is enabled.
    enabled: bool = True


class DropSite:
    """ This mixin class implements a drop site.  Any derived class must also
    derive QTreeWidgetItem.
    """

    def __init__(self):
        """ Initialise the instance. """

        self.setFlags(self.flags() | Qt.ItemFlag.ItemIsDropEnabled)

    def droppable(self, view):
        """ Return True if a view can be dropped. """

        raise NotImplementedError

    def drop(self, view):
        """ Handle the drop of a view. """

        raise NotImplementedError


# Each QTreeWidgetItem has a unique key used in the DND support.
_view_id = QTreeWidgetItem.ItemType.UserType


class APIView(QTreeWidgetItem):
    """ This class represents the view of an API in the editor. """

    def __init__(self, api, shell, parent, after=None):
        """ Initialise the view. """

        global _view_id
        view_id = _view_id
        _view_id += 1

        if after is parent:
            super().__init__(parent, None, view_id)
        elif after is None:
            super().__init__(parent, view_id)
        else:
            super().__init__(parent, after, view_id)

        self.api = api
        self.shell = shell

    def all_child_views(self):
        """ A generator for all the views's children. """

        for index in range(self.childCount()):
            yield self.child(index)

    def api_add(self, api):
        """ An API has been added. """

        # The order of views must match the order of APIs.
        index = self.api.content.find(api)
        after = self if index == 0 else self.child(index - 1)
        self.get_child_factory()(api, self.shell, self, after)

    def api_delete(self, api):
        """ An API has been deleted. """

        for view in self.all_child_views():
            if view.api is api:
                self.removeChild(view)
                break

    def api_as_str(self):
        """ Returns the API as a string for display purposes. """

        return adapt(self.api).as_str()

    def get_child_factory(self):
        """ Return the callable that will return a child instance. """

        raise NotImplementedError

    def get_menu(self, siblings):
        """ Return the list of context menu options or None if the view doesn't
        have a context menu.  Each element is a 3-tuple of the text of the
        option, the bound method that handles the option, and a flag that is
        True if the option should be enabled.
        """

        # This default implementation doesn't have a menu.
        return None


class ProjectView(APIView):
    """ This class implements a view of a project. """

    def __init__(self, shell, parent):
        """ Initialise the project view. """

        super().__init__(shell.project, shell, parent)

        self.root_module_updated()
        self.setExpanded(True)

        project = self.api

        # Progress will be reported against .sip files so count them all.
        nr_steps = 0
        for module in project.modules:
            nr_steps += len(module.content)

        # Display the progress dialog.
        progress = QProgressDialog("Building the GUI...", None, 0, nr_steps)
        progress.setWindowTitle(project.name)
        progress.setValue(0)

        so_far = 0
        for module in project.modules:
            ModuleView(module, self.shell, self)

            so_far += len(module.content)
            progress.setValue(so_far)
            QApplication.processEvents()

        self._sort()

    def get_menu(self, siblings):
        """ Return the list of context menu options. """

        if siblings:
            return None

        return [MenuOption("Properties...", self._handle_project_properties)]

    def module_add(self, module):
        """ Handle the addition of a module. """

        ModuleView(module, self.shell, self)
        self._sort()

    def module_delete(self, module):
        """ Handle the deletion of a module. """

        for view in self.all_child_views():
            if view.api is module:
                self.removeChild(view)
                break

    def module_rename(self, module):
        """ A module has been renamed. """

        for view in self.all_child_views():
            if view.api is module:
                view.setText(ApiEditor.NAME, module.name)
                break

        self._sort()

    def root_module_updated(self):
        """ The name of the root module has been updated. """

        name = self.shell.project.rootmodule
        if name == '':
            name = "Modules"

        self.setText(ApiEditor.NAME, name)

    def _handle_project_properties(self):
        """ Handle the project's properties. """

        dialog = ProjectPropertiesDialog(self.shell.project,
                "Project Properties", self.shell)

        if dialog.update():
            self.shell.dirty = True

    def _sort(self):
        """ Sort the modules. """

        self.sortChildren(ApiEditor.NAME, Qt.SortOrder.AscendingOrder)


class ModuleView(APIView, DropSite):
    """ This class implements a view of a module. """

    def __init__(self, module, shell, parent):
        """ Initialise the module view. """

        super().__init__(module, shell, parent)

        self.setText(ApiEditor.NAME, module.name)

        for sip_file in module.content:
            SipFileView(sip_file, self.shell, self)

    def droppable(self, view):
        """ Return True if a view can be dropped. """

        # We allow .sip files to be moved around anywhere.
        return isinstance(view, SipFileView)

    def drop(self, view):
        """ Handle the drop of a view. """

        # The .sip file is always placed at the top.
        sip_file = view.api
        dst_api = self.api.content
        src_api = view.parent().api.content

        src_api.remove(sip_file)
        dst_api.insert(0, sip_file)

        self.shell.dirty = True

    def get_child_factory(self):
        """ Return the callable that will return a child view. """

        return SipFileView

    def get_menu(self, siblings):
        """ Return the list of context menu options. """

        if siblings:
            return None

        return [MenuOption("Properties...", self._handle_module_properties)]

    def _handle_module_properties(self):
        """ Handle the module's properties. """

        dialog = ModulePropertiesDialog(self.api, "Module Properties",
                self.shell)

        if dialog.update():
            self.shell.dirty = True


class ContainerView(APIView, DropSite):
    """ This class implements a view of a potential container for code. """

    # The current external editor objects.
    editors = {}

    def __init__(self, container, shell, parent, after):
        """ Initialise the container view. """

        super().__init__(container, shell, parent, after)

        if hasattr(container, 'content'):
            for code in container.content:
                CodeView(code, self.shell, self)

    @classmethod
    def add_editor_option(cls, menu, name, handler, value, editor_id):
        """ Add the option that will invoke the external editor to a menu. """

        menu.append(
                MenuOption(name + '...', handler, checked=(value != ''),
                        enabled=(editor_id not in cls.editors)))

    def droppable(self, view):
        """ Return True if a view can be dropped. """

        # See if we are moving a sibling.
        if view.parent() is self.parent():
            return True

        # See if we are moving a child to the top.
        if view.parent() is self:
            return True

        return False

    def drop(self, view):
        """ Handle the drop of a view. """

        if view.parent() is self.parent():
            # We are moving a sibling after us.
            parent_content = self.parent().api.content
            api = view.api

            parent_content.remove(api)

            new_idx = parent_content.index(self.api) + 1
            if new_idx < len(parent_content):
                parent_content.insert(new_idx, api)
            else:
                parent_content.append(api)

            self.shell.dirty = True

            return

        if view.parent() is self:
            # Dropping a child is interpreted as moving it to the top.
            my_content = self.api.content
            api = view.api

            my_content.remove(api)
            my_content.insert(0, api)

            self.shell.dirty = True

            return

    def get_child_factory(self):
        """ Return the callable that will return a child instance. """

        return CodeView


class SipFileView(ContainerView):
    """ This class implements a view of a .sip file. """

    def __init__(self, sip_file, shell, parent, after=None):
        """ Initialise the .sip file view. """

        super().__init__(sip_file, shell, parent, after)

        self._targets = []

        self.setText(ApiEditor.NAME, sip_file.name)

    def droppable(self, view):
        """ Return True if a view can be dropped. """

        # See if we are moving another .sip file.
        if isinstance(view, SipFileView):
            return True

        # See if we are moving a child to the top.
        if view.parent() is self:
            return True

        return False

    def drop(self, view):
        """ Handle the drop of a view. """

        if isinstance(view, SipFileView):
            # We are moving another .sip file after us.  First remove it from
            # its container.
            view.parent().api.content.remove(view.api)

            # Now add it to our container.
            content = self.parent().api.content
            idx = content.index(self.api) + 1
            if idx < len(content):
                content.insert(idx, view.api)
            else:
                content.append(view.api)

            self.shell.dirty = True

            return

        if view.parent() is self:
            # Dropping a child is interpreted as moving it to the top.
            content = self.api.content

            content.remove(view.api)
            content.insert(0, view.api)

            self.shell.dirty = True

            return

    def get_menu(self, siblings):
        """ Return the list of context menu options. """

        if siblings:
            return None

        multiple_modules = (len(self.shell.project.modules) > 0)

        empty_sipfile = True
        for code in self.api.content:
            if code.status != 'ignored':
                empty_sipfile = False
                break

        menu = []

        menu.append(MenuOption("Hide Ignored", self._hideIgnoredSlot)),
        menu.append(MenuOption("Show Ignored", self._showIgnoredSlot)),
        menu.append(None)
        menu.append(
                MenuOption("Add manual code...",
                        self._handle_add_manual_code))
        menu.append(None)
        self.add_editor_option(menu, "%ExportedHeaderCode",
                self._exportedHeaderCodeSlot, self.api.exportedheadercode,
                'ehc')
        self.add_editor_option(menu, "%ModuleHeaderCode",
                self._moduleHeaderCodeSlot, self.api.moduleheadercode, 'mhc')
        self.add_editor_option(menu, "%ModuleCode", self._moduleCodeSlot,
                self.api.modulecode, 'moc')
        self.add_editor_option(menu, "%PreInitialisationCode",
                self._preInitCodeSlot, self.api.preinitcode, 'pric')
        self.add_editor_option(menu, "%InitialisationCode", self._initCodeSlot,
                self.api.initcode, 'ic')
        self.add_editor_option(menu, "%PostInitialisationCode",
                self._postInitCodeSlot, self.api.postinitcode, 'poic')
        menu.append(None)
        self.add_editor_option(menu, "%ExportedTypeHintCode",
                self._exportedTypeHintCodeSlot, self.api.exportedtypehintcode,
                'ethc')
        self.add_editor_option(menu, "%TypeHintCode", self._typeHintCodeSlot,
                self.api.typehintcode, 'thc')
        menu.append(None)
        menu.append(MenuOption("Move to...", self._handle_move,
                enabled=multiple_modules))
        menu.append(MenuOption("Delete", self._deleteFile,
                enabled=empty_sipfile))

        return menu

    def _handle_move(self):
        """ Move a .sip file to a different module. """

        src_module = self.parent().api

        dialog = MoveHeaderDialog(src_module, "Move Header File", self.shell)

        dst_module = dialog.get_destination_module()
        if dst_module is not None:
            # Mark as dirty before moving it.
            self.shell.dirty = True

            src_module.content.remove(self.api)
            dst_module.content.append(self.api)

    def _deleteFile(self):
        """ Delete an empty .sip file. """

        ans = QMessageBox.question(self.treeWidget(), "Delete header file",
                "Are you sure you want to delete this header file?",
                QMessageBox.StandardButton.Yes|QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No)

        if ans is QMessageBox.StandardButton.Yes:
            # Mark as dirty before removing it.
            self.shell.dirty = True

            self.parent().api.content.remove(self.api)

    def _handle_add_manual_code(self):
        """ Slot to handle the creation of manual code. """

        manual_code = ManualCode()

        dialog = ManualCodeDialog(manual_code, "Add Manual Code", self.shell)

        if dialog.update():
            self.api.content.insert(0, manual_code)
            self.shell.dirty = True

    def _exportedHeaderCodeSlot(self):
        """ Slot to handle %ExportedHeaderCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._exportedHeaderCodeDone)
        ed.edit(self.api.exportedheadercode, "%ExportedHeaderCode: " + self.api.name)
        self.editors['ehc'] = ed

    def _exportedHeaderCodeDone(self, text_changed, text):
        """ Slot to handle changed %ExportedHeaderCode. """

        if text_changed:
            self.api.exportedheadercode = text
            self.shell.dirty = True

        del self.editors['ehc']

    def _moduleHeaderCodeSlot(self):
        """ Slot to handle %ModuleHeaderCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._moduleHeaderCodeDone)
        ed.edit(self.api.moduleheadercode, "%ModuleHeaderCode: " + self.api.name)
        self.editors['mhc'] = ed

    def _moduleHeaderCodeDone(self, text_changed, text):
        """ Slot to handle changed %ModuleHeaderCode. """

        if text_changed:
            self.api.moduleheadercode = text
            self.shell.dirty = True

        del self.editors['mhc']

    def _moduleCodeSlot(self):
        """ Slot to handle %ModuleCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._moduleCodeDone)
        ed.edit(self.api.modulecode, "%ModuleCode: " + self.api.name)
        self.editors['moc'] = ed

    def _moduleCodeDone(self, text_changed, text):
        """ Slot to handle changed %ModuleCode. """

        if text_changed:
            self.api.modulecode = text
            self.shell.dirty = True

        del self.editors['moc']

    def _preInitCodeSlot(self):
        """ Slot to handle %PreInitialisationCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._preInitCodeDone)
        ed.edit(self.api.preinitcode, "%PreInitialisationCode: " + self.api.name)
        self.editors['pric'] = ed

    def _preInitCodeDone(self, text_changed, text):
        """ Slot to handle changed %PreInitialisationCode. """

        if text_changed:
            self.api.preinitcode = text
            self.shell.dirty = True

        del self.editors['pric']

    def _initCodeSlot(self):
        """ Slot to handle %InitialisationCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._initCodeDone)
        ed.edit(self.api.initcode, "%InitialisationCode: " + self.api.name)
        self.editors['ic'] = ed

    def _initCodeDone(self, text_changed, text):
        """ Slot to handle changed %InitialisationCode. """

        if text_changed:
            self.api.initcode = text
            self.shell.dirty = True

        del self.editors['ic']

    def _postInitCodeSlot(self):
        """ Slot to handle %PostInitialisationCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._postInitCodeDone)
        ed.edit(self.api.postinitcode, "%PostInitialisationCode: " + self.api.name)
        self.editors['poic'] = ed

    def _postInitCodeDone(self, text_changed, text):
        """ Slot to handle changed %PostInitialisationCode. """

        if text_changed:
            self.api.postinitcode = text
            self.shell.dirty = True

        del self.editors['poic']

    def _exportedTypeHintCodeSlot(self):
        """ Slot to handle %ExportedTypeHintCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._exportedTypeHintCodeDone)
        ed.edit(self.api.exportedtypehintcode, "%ExportedTypeHintCode: " + self.api.name)
        self.editors['ethc'] = ed

    def _exportedTypeHintCodeDone(self, text_changed, text):
        """ Slot to handle changed %ExportedTypeHintCode. """

        if text_changed:
            self.api.exportedtypehintcode = text
            self.shell.dirty = True

        del self.editors['ethc']

    def _typeHintCodeSlot(self):
        """ Slot to handle %TypeHintCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._typeHintCodeDone)
        ed.edit(self.api.typehintcode, "%TypeHintCode: " + self.api.name)
        self.editors['thc'] = ed

    def _typeHintCodeDone(self, text_changed, text):
        """ Slot to handle changed %TypeHintCode. """

        if text_changed:
            self.api.typehintcode = text
            self.shell.dirty = True

        del self.editors['thc']

    def _hideIgnoredSlot(self):
        """ Hide all ignored scope elements. """

        self._setIgnoredVisibility(False)

    def _showIgnoredSlot(self):
        """ Show all ignored scope elements. """

        self._setIgnoredVisibility(True)

    def _setIgnoredVisibility(self, visible):
        """ Set the visibility of all ignored scope elements. """

        it = QTreeWidgetItemIterator(self)
        itm = it.value()

        while itm:
            if isinstance(itm, CodeView) and itm.code.status == "ignored":
                itm.setHidden(not visible)

            it += 1
            itm = it.value()


class ArgumentView(APIView):
    """ This class implements a view of a function argument. """

    def __init__(self, argument, shell, parent):
        """ Initialise the argument view. """

        super().__init__(argument, shell, parent)

        self.setFlags(Qt.ItemFlag.ItemIsEnabled)

        self.draw_name()

        # Stash this so that an argument can re-draw itself.
        argument._view = self

    def draw_name(self):
        """ Draw the name column. """

        self.setText(ApiEditor.NAME, self.api_as_str())

    def get_menu(self, siblings):
        """ Return the list of context menu options. """

        if siblings:
            return None

        return [MenuOption("Properties...", self._handle_argument_properties)]

    def _handle_argument_properties(self):
        """ Slot to handle the argument's properties. """

        dialog = ArgumentPropertiesDialog(self.api, "Argument Properties",
                self.shell)

        if dialog.update():
            self.shell.dirty = True

            self.draw_name()
            self.parent().draw_name()
            self.parent().draw_status()


class CodeView(ContainerView):
    """ This class implements a view of code in a .sip file. """

    def __init__(self, code, shell, parent, after=None):
        """ Initialise the code view. """

        super().__init__(code, shell, parent, after)

        self._targets = []

        self.draw_name()
        self._draw_access()
        self.draw_status()
        self.draw_versions()

        if code.status == 'ignored':
            self.setHidden(True)

        # Create any children.
        if hasattr(code, 'args'):
            for arg in code.args:
                ArgumentView(arg, self.shell, self)

    def draw_name(self):
        """ Update the item's name. """

        self.setText(ApiEditor.NAME, self.api_as_str())

    def _draw_access(self):
        """ Update the item's access. """

        # Not everything has an access specifier.
        try:
            access = self.api.access
        except AttributeError:
            access = ''

        self.setText(ApiEditor.ACCESS, access)

    def _has_unnamed_args(self):
        """ Returns ``True`` if the code has unnamed arguments. """

        # These types don't use named arguments.
        if isinstance(self.api, (OperatorMethod, OperatorCast, OperatorFunction)):
            return False

        # Ignore private items.
        try:
            private = (self.api.access == 'private')
        except AttributeError:
            private = False

        if private:
            return False

        for arg in self.api.args:
            if arg.unnamed and arg.default != '':
                return True

        return False

    def draw_status(self):
        """ Update the item's status. """

        status = []

        expand = False
        s = self.api.status

        if s == 'removed':
            text = "Removed"
            expand = True
        elif s == 'todo':
            text = "Todo"
            expand = True
        elif s == 'unknown':
            text = "Unchecked"
            expand = True
        elif s == 'ignored':
            text = "Ignored"
        else:
            text = ''

        if text:
            status.append(text)

        if s != 'ignored' and hasattr(self.api, 'args') and self._has_unnamed_args():
            status.append("Unnamed arguments")
            expand = True

        self.setText(ApiEditor.STATUS, ", ".join(status))

        if expand:
            parent = self.parent()
            while parent is not None:
                if isinstance(parent, CodeView):
                    if parent.api.status == 'ignored':
                        break

                parent.setExpanded(True)
                parent = parent.parent()

    def draw_versions(self):
        """ Update the item's versions. """

        self.setText(ApiEditor.VERSIONS,
                adapt(self.api, Tagged).versions_as_str())

    def get_menu(self, siblings):
        """ Return the list of context menu options. """

        project = self.shell.project

        # Save the list of targets for the menu action, including this one.
        self._targets = siblings[:]
        self._targets.append(self)

        menu = []

        # Handle the workflow.
        menu.append(
                MenuOption("Checked", self._setStatusChecked,
                        checked=(self.api.status == '')))
        menu.append(
                MenuOption("Todo", self._setStatusTodo,
                        checked=(self.api.status == 'todo')))
        menu.append(
                MenuOption("Unchecked", self._setStatusUnchecked,
                        checked=(self.api.status == 'unknown')))
        menu.append(
                MenuOption("Ignored", self._setStatusIgnored,
                        checked=(self.api.status == 'ignored')))

        menu.append(None)
        self.add_editor_option(menu, "Comments", self._handle_comments,
                self.api.comments, 'comments')

        # Handle the access specifiers.
        if isinstance(self.api, Access):
                menu.append(None)
                menu.append(
                        MenuOption("public", self._setAccessPublic,
                                checked=(self.api.access == '')))
                menu.append(
                        MenuOption("public slots", self._setAccessPublicSlots,
                                checked=(self.api.access == 'public slots')))
                menu.append(
                        MenuOption("protected", self._setAccessProtected,
                                checked=(self.api.access == 'protected')))
                menu.append(
                        MenuOption("protected slots",
                                self._setAccessProtectedSlots,
                                checked=(self.api.access == 'protected slots')))
                menu.append(
                        MenuOption("private", self._setAccessPrivate,
                                checked=(self.api.access == 'private')))
                menu.append(
                        MenuOption("private slots",
                                self._setAccessPrivateSlots,
                                checked=(self.api.access == 'private slots')))
                menu.append(
                        MenuOption("signals", self._setAccessSignals,
                                checked=(self.api.access == 'signals')))

        if siblings:
            if project.versions:
                menu.append(None)
                menu.append(MenuOption("Versions...", self._handle_versions))

            menu.append(None)
            menu.append(
                    MenuOption("Delete", self._deleteCode,
                            enabled=(not self.editors)))

            return menu

        # Handle the manual code part of the menu.
        menu.append(None)
        menu.append(
                MenuOption("Add manual code...", self._handle_add_manual_code))

        # See what extra menu items are needed.
        thcslot = False
        thicslot = False
        tcslot = False
        cttcslot = False
        cftcslot = False
        mcslot = False
        vccslot = False
        fcslot = False
        sccslot = False
        acslot = False
        gcslot = False
        scslot = False
        gctcslot = False
        gcccslot = False
        bigetbslot = False
        birelbslot = False
        birbslot = False
        biwbslot = False
        biscslot = False
        bicbslot = False
        pickslot = False
        xaslot = False
        properties_handler = None
        dsslot = None

        if isinstance(self.api, ManualCode):
            menu.append(
                    MenuOption("Modify manual code...",
                            self._handle_modify_manual_code))
            menu.append(
                    MenuOption("Modify manual code body...",
                            self._bodyManualCode,
                            checked=(self.api.body != ''),
                            enabled=('mcb' not in self.editors)))

            mcslot = True
            properties_handler = self._handle_callable_properties
            dsslot = True
        elif isinstance(self.api, Namespace):
            properties_handler = self._handle_namespace_properties
        elif isinstance(self.api, OpaqueClass):
            properties_handler = self._handle_opaque_class_properties
        elif isinstance(self.api, Namespace):
            thcslot = True
        elif isinstance(self.api, Class):
            thcslot = True
            thicslot = True
            tcslot = True
            cttcslot = True
            cftcslot = True
            fcslot = True
            sccslot = True
            gctcslot = True
            gcccslot = True
            bigetbslot = True
            birelbslot = True
            birbslot = True
            biwbslot = True
            biscslot = True
            bicbslot = True
            pickslot = True
            xaslot = True
            properties_handler = self._handle_class_properties
            dsslot = True
        elif isinstance(self.api, Constructor):
            mcslot = True
            properties_handler = self._handle_callable_properties
            dsslot = True
        elif isinstance(self.api, Destructor):
            mcslot = True
            vccslot = True
            properties_handler = self._handle_callable_properties
        elif isinstance(self.api, OperatorCast):
            mcslot = True
            properties_handler = self._handle_callable_properties
        elif isinstance(self.api, Method):
            mcslot = True

            if self.api.virtual:
                vccslot = True

            properties_handler = self._handle_callable_properties
            dsslot = True
        elif isinstance(self.api, OperatorMethod):
            mcslot = True

            if self.api.virtual:
                vccslot = True

            properties_handler = self._handle_callable_properties
        elif isinstance(self.api, Function):
            mcslot = True
            properties_handler = self._handle_callable_properties
            dsslot = True
        elif isinstance(self.api, OperatorFunction):
            mcslot = True
            properties_handler = self._handle_callable_properties
        elif isinstance(self.api, Typedef):
            properties_handler = self._handle_typedef_properties
            dsslot = True
        elif isinstance(self.api, Variable):
            acslot = True
            gcslot = True
            scslot = True
            properties_handler = self._handle_variable_properties
        elif isinstance(self.api, Enum):
            properties_handler = self._handle_enum_properties
        elif isinstance(self.api, EnumValue):
            properties_handler = self._handle_enum_member_properties

        if thcslot or thicslot or tcslot or cttcslot or cftcslot or fcslot or sccslot or mcslot or vccslot or acslot or gcslot or scslot or gctcslot or gcccslot or bigetbslot or birelbslot or birbslot or biwbslot or biscslot or bicbslot or pickslot or xaslot or dsslot:
            menu.append(None)

            if thcslot:
                self.add_editor_option(menu, '%TypeHeaderCode',
                        self._typeheaderCodeSlot, self.api.typeheadercode,
                        'thc')

            if tcslot:
                self.add_editor_option(menu, '%TypeCode', self._typeCodeSlot,
                        self.api.typecode, 'tc')

            if thicslot:
                self.add_editor_option(menu, '%TypeHintCode',
                        self._typehintCodeSlot, self.api.typehintcode, 'thic')

            if fcslot:
                self.add_editor_option(menu, '%FinalisationCode',
                        self._finalCodeSlot, self.api.finalisationcode, 'fc')

            if sccslot:
                self.add_editor_option(menu, '%ConvertToSubClassCode',
                        self._subclassCodeSlot, self.api.subclasscode, 'scc')

            if cttcslot:
                self.add_editor_option(menu, '%ConvertToTypeCode',
                        self._convToTypeCodeSlot, self.api.convtotypecode,
                        'cttc')

            if cftcslot:
                self.add_editor_option(menu, '%ConvertFromTypeCode',
                        self._convFromTypeCodeSlot, self.api.convfromtypecode,
                        'cftc')

            if mcslot:
                self.add_editor_option(menu, '%MethodCode',
                        self._methodCodeSlot, self.api.methcode, 'mc')

            if vccslot:
                self.add_editor_option(menu, '%VirtualCatcherCode',
                        self._virtualCatcherCodeSlot, self.api.virtcode, 'vcc')

            if acslot:
                self.add_editor_option(menu, '%AccessCode',
                        self._accessCodeSlot, self.api.accesscode, 'ac')

            if gcslot:
                self.add_editor_option(menu, '%GetCode', self._getCodeSlot,
                        self.api.getcode, 'gc')

            if scslot:
                self.add_editor_option(menu, '%SetCode', self._setCodeSlot,
                        self.api.setcode, 'sc')

            if gctcslot:
                self.add_editor_option(menu, '%GCTraverseCode',
                        self._gcTraverseCodeSlot, self.api.gctraversecode,
                        'gctc')

            if gcccslot:
                self.add_editor_option(menu, '%GCClearCode',
                        self._gcClearCodeSlot, self.api.gcclearcode, 'gccc')

            if bigetbslot:
                self.add_editor_option(menu, '%BIGetBufferCode',
                        self._biGetBufCodeSlot, self.api.bigetbufcode,
                        'bigetb')

            if birelbslot:
                self.add_editor_option(menu, '%BIReleaseBufferCode',
                        self._biRelBufCodeSlot, self.api.birelbufcode,
                        'birelb')

            if birbslot:
                self.add_editor_option(menu, '%BIGetReadBufferCode',
                        self._biReadBufCodeSlot, self.api.bireadbufcode,
                        'birb')

            if biwbslot:
                self.add_editor_option(menu, '%BIGetWriteBufferCode',
                        self._biWriteBufCodeSlot, self.api.biwritebufcode,
                        'biwb')

            if biscslot:
                self.add_editor_option(menu, '%BIGetSegCountCode',
                        self._biSegCountCodeSlot, self.api.bisegcountcode,
                        'bisc')

            if bicbslot:
                self.add_editor_option(menu, '%BIGetCharBufferCode',
                        self._biCharBufCodeSlot, self.api.bicharbufcode,
                        'bicb')

            if pickslot:
                self.add_editor_option(menu, '%PickleCode',
                        self._pickleCodeSlot, self.api.picklecode, 'pick')

            if dsslot:
                self.add_editor_option(menu, '%Docstring', self._docstringSlot,
                        self.api.docstring, 'ds')

        if isinstance(self.api, (Constructor, Function, Method)):
            menu.append(None)
            menu.append(
                    MenuOption("Accept all argument names", self._acceptNames))

        if project.versions or project.platforms or project.features or project.externalfeatures:
            menu.append(None)

            if project.versions:
                menu.append(MenuOption("Versions...", self._handle_versions))

            if project.platforms:
                menu.append(
                        MenuOption("Platform Tags...", self._handle_platforms,
                                checked=bool(self.api.platforms)))

            if project.features or project.externalfeatures:
                menu.append(
                        MenuOption("Feature Tags...", self._handle_features,
                                checked=bool(self.api.features)))

        if properties_handler is not None:
            menu.append(MenuOption("Properties...", properties_handler))

        menu.append(None)
        menu.append(
                MenuOption("Delete", self._deleteCode,
                        enabled=(not self.editors)))

        return menu

    def _acceptNames(self):
        """ Accept all argument names. """

        updated = False

        for arg in self.api.args:
            if arg.unnamed and arg.default != '':
                arg.unnamed = False
                arg._view.draw_name()
                updated = True

        if updated:
            self.draw_name()
            self.draw_status()
            self.shell.dirty = True

    def _handle_comments(self):
        """ Slot to handle comments. """

        name = self.api.precis if isinstance(self.api, ManualCode) else self.api.name

        ed = ExternalEditor()
        ed.editDone.connect(self._handle_comments_done)
        ed.edit(self.api.comments, "Comments: " + name)
        self.editors['comments'] = ed

    def _handle_comments_done(self, text_changed, text):
        """ Slot to handle changed comments. """

        if text_changed:
            self.api.comments = text
            self.shell.dirty = True

        del self.editors['comments']

    def _handle_add_manual_code(self):
        """ Slot to handle the addition of manual code. """

        manual_code = ManualCode()

        dialog = ManualCodeDialog(manual_code, "Add Manual Code", self.shell)

        if dialog.update():
            parent_content = self.parent().api.content
            parent_content.insert(parent_content.index(self.api) + 1,
                    manual_code)
            self.shell.dirty = True

    def _handle_modify_manual_code(self):
        """ Slot to handle the update of the manual code. """

        dialog = ManualCodeDialog(self.api, "Modify Manual Code", self.shell)

        if dialog.update():
            self.shell.dirty = True

            self.draw_name()

    def _bodyManualCode(self):
        """ Slot to handle the update of the manual code body. """

        ed = ExternalEditor()
        ed.editDone.connect(self._mcBodyDone)
        ed.edit(self.api.body, "Manual Code: " + self.api.precis)
        self.editors['mcb'] = ed

    def _mcBodyDone(self, text_changed, text):
        """ Slot to handle changed manual code body. """

        if text_changed:
            self.api.body = text
            self.shell.dirty = True

        del self.editors['mcb']

    def _deleteCode(self):
        """ Slot to handle the deletion of one or more code items. """

        what = "this part" if len(self._targets) == 1 else "these parts"
        ans = QMessageBox.question(self.treeWidget(), "Delete Code",
                "Are you sure you want to delete {0} of the API?".format(what),
                QMessageBox.StandardButton.Yes|QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No)

        if ans is QMessageBox.StandardButton.Yes:
            # Mark as dirty before removing them.
            self.shell.dirty = True

            for target in self._targets:
                self.parent().api.content.remove(target.api)

    def _accessCodeSlot(self):
        """ Slot to handle %AccessCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._accessCodeDone)
        ed.edit(self.api.accesscode, "%AccessCode: " + self.api_as_str())
        self.editors['ac'] = ed

    def _accessCodeDone(self, text_changed, text):
        """ Slot to handle changed %AccessCode. """

        if text_changed:
            self.api.accesscode = text
            self.shell.dirty = True

        del self.editors['ac']

    def _getCodeSlot(self):
        """ Slot to handle %GetCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._getCodeDone)
        ed.edit(self.api.getcode, "%GetCode: " + self.api_as_str())
        self.editors['gc'] = ed

    def _getCodeDone(self, text_changed, text):
        """ Slot to handle changed %GetCode. """

        if text_changed:
            self.api.getcode = text
            self.shell.dirty = True

        del self.editors['gc']

    def _setCodeSlot(self):
        """ Slot to handle %SetCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._setCodeDone)
        ed.edit(self.api.setcode, "%SetCode: " + self.api_as_str())
        self.editors['sc'] = ed

    def _setCodeDone(self, text_changed, text):
        """ Slot to handle changed %SetCode. """

        if text_changed:
            self.api.setcode = text
            self.shell.dirty = True

        del self.editors['sc']

    def _typehintCodeSlot(self):
        """ Slot to handle %TypeHintCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._typehintCodeDone)
        ed.edit(self.api.typeheadercode, "%TypeHintCode: " + self.api_as_str())
        self.editors['thic'] = ed

    def _typehintCodeDone(self, text_changed, text):
        """ Slot to handle changed %TypeHintCode. """

        if text_changed:
            self.api.typehintcode = text
            self.shell.dirty = True

        del self.editors['thic']

    def _typeheaderCodeSlot(self):
        """ Slot to handle %TypeHeaderCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._typeheaderCodeDone)
        ed.edit(self.api.typeheadercode,
                "%TypeHeaderCode: " + self.api_as_str())
        self.editors['thc'] = ed

    def _typeheaderCodeDone(self, text_changed, text):
        """ Slot to handle changed %TypeHeaderCode. """

        if text_changed:
            self.api.typeheadercode = text
            self.shell.dirty = True

        del self.editors['thc']

    def _typeCodeSlot(self):
        """ Slot to handle %TypeCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._typeCodeDone)
        ed.edit(self.api.typecode, "%TypeCode: " + self.api_as_str())
        self.editors['tc'] = ed

    def _typeCodeDone(self, text_changed, text):
        """ Slot to handle changed %TypeCode. """

        if text_changed:
            self.api.typecode = text
            self.shell.dirty = True

        del self.editors['tc']

    def _convToTypeCodeSlot(self):
        """ Slot to handle %ConvertToTypeCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._convToTypeCodeDone)
        ed.edit(self.api.convtotypecode,
                "%ConvertToTypeCode: " + self.api_as_str())
        self.editors['cttc'] = ed

    def _convToTypeCodeDone(self, text_changed, text):
        """ Slot to handle changed %ConvertToTypeCode. """

        if text_changed:
            self.api.convtotypecode = text
            self.shell.dirty = True

        del self.editors['cttc']

    def _convFromTypeCodeSlot(self):
        """ Slot to handle %ConvertFromTypeCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._convFromTypeCodeDone)
        ed.edit(self.api.convfromtypecode,
                "%ConvertFromTypeCode: " + self.api_as_str())
        self.editors['cftc'] = ed

    def _convFromTypeCodeDone(self, text_changed, text):
        """ Slot to handle changed %ConvertFromTypeCode. """

        if text_changed:
            self.api.convfromtypecode = text
            self.shell.dirty = True

        del self.editors['cftc']

    def _gcTraverseCodeSlot(self):
        """ Slot to handle %GCTraverseCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._gcTraverseCodeDone)
        ed.edit(self.api.gctraversecode,
                "%GCTraverseCode: " + self.api_as_str())
        self.editors['gctc'] = ed

    def _gcTraverseCodeDone(self, text_changed, text):
        """ Slot to handle changed %GCTraverseCode. """

        if text_changed:
            self.api.gctraversecode = text
            self.shell.dirty = True

        del self.editors['gctc']

    def _gcClearCodeSlot(self):
        """ Slot to handle %GCClearCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._gcClearCodeDone)
        ed.edit(self.api.gcclearcode, "%GCClearCode: " + self.api_as_str())
        self.editors['gccc'] = ed

    def _gcClearCodeDone(self, text_changed, text):
        """ Slot to handle changed %GCClearCode. """

        if text_changed:
            self.api.gcclearcode = text
            self.shell.dirty = True

        del self.editors['gccc']

    def _biGetBufCodeSlot(self):
        """ Slot to handle %BIGetBufferCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._biGetBufCodeDone)
        ed.edit(self.api.bigetbufcode,
                "%BIGetBufferCode: " + self.api_as_str())
        self.editors['bigetb'] = ed

    def _biGetBufCodeDone(self, text_changed, text):
        """ Slot to handle changed %BIGetBufferCode. """

        if text_changed:
            self.api.bigetbufcode = text
            self.shell.dirty = True

        del self.editors['bigetb']

    def _biRelBufCodeSlot(self):
        """ Slot to handle %BIReleaseBufferCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._biRelBufCodeDone)
        ed.edit(self.api.birelbufcode,
                "%BIReleaseBufferCode: " + self.api_as_str())
        self.editors['birelb'] = ed

    def _biRelBufCodeDone(self, text_changed, text):
        """ Slot to handle changed %BIReleaseBufferCode. """

        if text_changed:
            self.api.birelbufcode = text
            self.shell.dirty = True

        del self.editors['birelb']

    def _biReadBufCodeSlot(self):
        """ Slot to handle %BIGetReadBufferCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._biReadBufCodeDone)
        ed.edit(self.api.bireadbufcode,
                "%BIGetReadBufferCode: " + self.api_as_str())
        self.editors['birb'] = ed

    def _biReadBufCodeDone(self, text_changed, text):
        """ Slot to handle changed %BIGetReadBufferCode. """

        if text_changed:
            self.api.bireadbufcode = text
            self.shell.dirty = True

        del self.editors['birb']

    def _biWriteBufCodeSlot(self):
        """ Slot to handle %BIGetWriteBufferCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._biWriteBufCodeDone)
        ed.edit(self.api.biwritebufcode,
                "%BIGetWriteBufferCode: " + self.api_as_str())
        self.editors['biwb'] = ed

    def _biWriteBufCodeDone(self, text_changed, text):
        """ Slot to handle changed %BIGetWriteBufferCode. """

        if text_changed:
            self.api.biwritebufcode = text
            self.shell.dirty = True

        del self.editors['biwb']

    def _biSegCountCodeSlot(self):
        """ Slot to handle %BIGetSegCountCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._biSegCountCodeDone)
        ed.edit(self.api.bisegcountcode,
                "%BIGetSegCountCode: " + self.api_as_str())
        self.editors['bisc'] = ed

    def _biSegCountCodeDone(self, text_changed, text):
        """ Slot to handle changed %BIGetSegCountCode. """

        if text_changed:
            self.api.bisegcountcode = text
            self.shell.dirty = True

        del self.editors['bisc']

    def _biCharBufCodeSlot(self):
        """ Slot to handle %BIGetCharBufferCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._biCharBufCodeDone)
        ed.edit(self.api.bicharbufcode,
                "%BIGetCharBufferCode: " + self.api_as_str())
        self.editors['bicb'] = ed

    def _biCharBufCodeDone(self, text_changed, text):
        """ Slot to handle changed %BIGetCharBufferCode. """

        if text_changed:
            self.api.bicharbufcode = text
            self.shell.dirty = True

        del self.editors['bicb']

    def _pickleCodeSlot(self):
        """ Slot to handle %PickleCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._pickleCodeDone)
        ed.edit(self.api.picklecode, "%PickleCode: " + self.api_as_str())
        self.editors['pick'] = ed

    def _pickleCodeDone(self, text_changed, text):
        """ Slot to handle changed %PickleCode. """

        if text_changed:
            self.api.picklecode = text
            self.shell.dirty = True

        del self.editors['pick']

    def _finalCodeSlot(self):
        """ Slot to handle %FinalisationCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._finalCodeDone)
        ed.edit(self.api.finalisationcode,
                "%FinalisationCode: " + self.api_as_str())
        self.editors['fc'] = ed

    def _finalCodeDone(self, text_changed, text):
        """ Slot to handle changed %FinalisationCode. """

        if text_changed:
            self.api.finalisationcode = text
            self.shell.dirty = True

        del self.editors['fc']

    def _subclassCodeSlot(self):
        """ Slot to handle %ConvertToSubClassCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._subclassCodeDone)
        ed.edit(self.api.subclasscode,
                "%ConvertToSubClassCode: " + self.api_as_str())
        self.editors['scc'] = ed

    def _subclassCodeDone(self, text_changed, text):
        """ Slot to handle changed %ConvertToSubClassCode. """

        if text_changed:
            self.api.subclasscode = text
            self.shell.dirty = True

        del self.editors['scc']

    def _docstringSlot(self):
        """ Slot to handle %Docstring. """

        ed = ExternalEditor()
        ed.editDone.connect(self._docstringDone)
        ed.edit(self.api.docstring, "%Docstring: " + self.api_as_str())
        self.editors['ds'] = ed

    def _docstringDone(self, text_changed, text):
        """ Slot to handle changed %Docstring. """

        if text_changed:
            self.api.docstring = text
            self.shell.dirty = True

        del self.editors['ds']

    def _methodCodeSlot(self):
        """ Slot to handle %MethodCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._methodCodeDone)
        ed.edit(self.api.methcode, "%MethodCode: " + self.api_as_str())
        self.editors['mc'] = ed

    def _methodCodeDone(self, text_changed, text):
        """ Slot to handle changed %MethodCode. """

        if text_changed:
            self.api.methcode = text
            self.shell.dirty = True

        del self.editors['mc']

    def _virtualCatcherCodeSlot(self):
        """ Slot to handle %VirtualCatcherCode. """

        ed = ExternalEditor()
        ed.editDone.connect(self._virtualCatcherCodeDone)
        ed.edit(self.api.virtcode, "%VirtualCatcherCode: " + self.api_as_str())
        self.editors['vcc'] = ed

    def _virtualCatcherCodeDone(self, text_changed, text):
        """ Slot to handle changed %VirtualCatcherCode. """

        if text_changed:
            self.api.virtcode = text
            self.shell.dirty = True

        del self.editors['vcc']

    def _handle_versions(self):
        """ Slot to handle the versions. """

        dialog = VersionsDialog(self.api, "Versions", self.shell)

        if dialog.update():
            # Apply the version range to all targets.
            for view in self._targets:
                if view.api is not self.api:
                    view.api.versions = list(self.api.versions)

            self.shell.dirty = True

    def _handle_platforms(self):
        """ Slot to handle the platforms. """

        dialog = PlatformsDialog(self.api, "Platform Tags", self.shell)

        if dialog.update():
            self.shell.dirty = True

    def _handle_features(self):
        """ Slot to handle the features. """

        dialog = FeaturesDialog(self.api, "Feature Tags", self.shell)

        if dialog.update():
            self.shell.dirty = True

    def _handle_namespace_properties(self):
        """ Slot to handle the properties for namespaces. """

        dialog = NamespacePropertiesDialog(self.api, "Namespace Properties",
                self.shell)

        if dialog.update():
            self.shell.dirty = True
            self.setText(ApiEditor.NAME, self.api_as_str())

    def _handle_opaque_class_properties(self):
        """ Slot to handle the properties for opaque classes. """

        dialog = OpaqueClassPropertiesDialog(self.api,
                "Opaque Class Properties", self.shell)

        if dialog.update():
            self.shell.dirty = True
            self.setText(ApiEditor.NAME, self.api_as_str())

    def _handle_class_properties(self):
        """ Slot to handle the properties for classes. """

        dialog = ClassPropertiesDialog(self.api, "Class Properties",
                self.shell)

        if dialog.update():
            self.shell.dirty = True
            self.setText(ApiEditor.NAME, self.api_as_str())

    def _handle_callable_properties(self):
        """ Slot to handle the properties for callables. """

        dialog = CallablePropertiesDialog(self.api, 'Placeholder', self.shell)

        if dialog.update():
            self.shell.dirty = True
            self.setText(ApiEditor.NAME, self.api_as_str())

    def _handle_typedef_properties(self):
        """ Slot to handle the properties for typedefs. """

        dialog = TypedefPropertiesDialog(self.api, "Typedef Properties",
                self.shell)

        if dialog.update():
            self.shell.dirty = True
            self.setText(ApiEditor.NAME, self.api_as_str())

    def _handle_variable_properties(self):
        """ Slot to handle the properties for variables. """

        dialog = VariablePropertiesDialog(self.api, "Variable Properties",
                self.shell)

        if dialog.update():
            self.shell.dirty = True
            self.setText(ApiEditor.NAME, self.api_as_str())

    def _handle_enum_properties(self):
        """ Slot to handle the properties for enums. """

        dialog = EnumPropertiesDialog(self.api, "Enum Properties", self.shell)

        if dialog.update():
            self.shell.dirty = True
            self.setText(ApiEditor.NAME, self.api_as_str())

    def _handle_enum_member_properties(self):
        """ Slot to handle the properties for enum members. """

        dialog = EnumMemberPropertiesDialog(self.api,
                "Enum Member Properties", self.shell)

        if dialog.update():
            self.shell.dirty = True
            self.setText(ApiEditor.NAME, self.api_as_str())

    def _setStatusChecked(self):
        """ Slot to handle the status being set to checked. """

        self._setStatus('')

    def _setStatusTodo(self):
        """ Slot to handle the status being set to todo. """

        self._setStatus('todo')

    def _setStatusUnchecked(self):
        """ Slot to handle the status being set to unchecked. """

        self._setStatus('unknown')

    def _setStatusIgnored(self):
        """ Slot to handle the status being set to ignored. """

        self._setStatus('ignored')

    def _setStatus(self, new):
        """ Set the status column. """

        for view in self._targets:
            if view.api.status != new:
                view.api.status = new
                self.shell.dirty = True

                # FIXME: Observe the status attribute.
                view.draw_status()

    def _setAccessPublic(self):
        """ Slot to handle the access being set to public. """

        self._setAccess('')

    def _setAccessPublicSlots(self):
        """ Slot to handle the access being set to public slots. """

        self._setAccess("public slots")

    def _setAccessProtected(self):
        """ Slot to handle the access being set to protected. """

        self._setAccess("protected")

    def _setAccessProtectedSlots(self):
        """ Slot to handle the access being set to protected slots. """

        self._setAccess("protected slots")

    def _setAccessPrivate(self):
        """ Slot to handle the access being set to private. """

        self._setAccess("private")

    def _setAccessPrivateSlots(self):
        """ Slot to handle the access being set to private slots. """

        self._setAccess("private slots")

    def _setAccessSignals(self):
        """ Slot to handle the access being set to signals. """

        self._setAccess("signals")

    def _setAccess(self, new):
        """ Set the access column. """

        for view in self._targets:
            if view.api.access != new:
                view.api.access = new
                self.shell.dirty = True

                # FIXME: Observe the access attribute.
                view._draw_access()



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass, field

from .version_range import VersionRange


@dataclass
class Tagged:
    """ This class is a mixin for APIs that are subject to tags (i.e. SIP's %If
    statement).
    """

    # The optional list of logically or-ed features that the API item is
    # limited to.  A feature may be preceded by "!" to indicate the logical
    # inverse.
    features: list[str] = field(default_factory=list)

    # The optional list of logically or-ed platforms that the API item is
    # limited to.  Note that SIP supports inverting a platform, as does the
    # MetaSIP platforms tool, but the MetaSIP UI for picking a platform doesn't
    # yet.
    platforms: list[str] = field(default_factory=list)

    # The optional list of logically and-ed version ranges.  Note that SIP
    # supports multiple ranges, as does the MetaSIP versions tool, but the
    # MetaSIP UI for picking a version range doesn't yet.
    versions: list[VersionRange] = field(default_factory=list)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from .access import Access
from .annos import Annos
from .argument import Argument
from .callable import Callable
from .code import Code
from .code_container import CodeContainer
from .constructor import Constructor
from .destructor import Destructor
from .docstring import Docstring
from .enum import Enum
from .enum_value import EnumValue
from .extended_access import ExtendedAccess
from .function import Function
from .header_directory import HeaderDirectory
from .header_file import HeaderFile
from .header_file_version import HeaderFileVersion
from .klass import Class
from .manual_code import ManualCode
from .method import Method
from .module import Module
from .namespace import Namespace
from .opaque_class import OpaqueClass
from .operator_cast import OperatorCast
from .operator_function import OperatorFunction
from .operator_method import OperatorMethod
from .platform import Platform
from .project import Project
from .sip_file import SipFile
from .tagged import Tagged
from .typedef import Typedef
from .variable import Variable
from .version_range import VersionRange
from .workflow import Workflow

from .project_version import MinimumProjectVersion, ProjectVersion



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtCore import pyqtSignal, QEvent, QFile, QObject, QTextStream
from PyQt6.QtWidgets import (QDialog, QDialogButtonBox, QFileDialog,
        QHBoxLayout, QMessageBox, QPushButton, QVBoxLayout)
from PyQt6.Qsci import QsciLexerCPP, QsciScintilla


class ExternalEditor(QObject):
    """
    This class represents an external editor.
    """

    editDone = pyqtSignal(bool, str)

    def edit(self, text, caption):
        """
        Start the external editor to edit some text.  The editDone() signal is
        emitted when the edit is finshed.  The signal passes a boolean that is
        set if the text has been modified and the text itself.

        text is the text to edit.
        caption is the window caption.
        """
        # Create the editor as a dialog.
        layout = QVBoxLayout()

        top_buttons = QHBoxLayout()

        read_file = QPushButton("Insert file", clicked=self._insert_file)
        top_buttons.addWidget(read_file)

        top_buttons.addStretch()
        layout.addLayout(top_buttons)

        self._editor = ed = QsciScintilla()

        ed.setUtf8(True)
        ed.setAutoIndent(True)
        ed.setIndentationWidth(4)
        ed.setIndentationsUseTabs(False)
        ed.setFolding(QsciScintilla.FoldStyle.PlainFoldStyle)
        ed.setEdgeColumn(80)
        ed.setEdgeMode(QsciScintilla.EdgeMode.EdgeLine)

        lexer = QsciLexerCPP(ed)
        ed.setLexer(lexer)

        # Use a mono-spaced font.
        s = 0
        while lexer.description(s) != '':
            f = lexer.font(s)
            f.setFamily('Courier')
            lexer.setFont(f)

            s += 1

        layout.addWidget(ed)

        bbox = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok|QDialogButtonBox.StandardButton.Cancel)
        bbox.accepted.connect(self._on_accepted)
        bbox.rejected.connect(self._on_rejected)
        layout.addWidget(bbox)

        self._dialog = dlg = QDialog(windowTitle=caption,
                objectName='metasip.api_editor.external_editor')
        dlg.setLayout(layout)

        # Load the text.
        ed.setText(text)

        dlg.installEventFilter(self)
        dlg.show()
        dlg.raise_()

    def eventFilter(self, obj, event):
        """ Reimplemented to handle any dialog close events. """

        if event.type() is QEvent.Type.Close:
            self._on_rejected()

        return super().eventFilter(obj, event)

    def _insert_file(self):
        """
        Invoked to insert the contents of a file.
        """
        fn, _ = QFileDialog.getOpenFileName(self._dialog, "Insert File")

        if fn == '':
            return

        f = QFile(fn)

        if not f.open(f.ReadOnly):
            QMessageBox.warning(self._dialog, "Insert File",
                    "Unable to open file")
            return

        ts = QTextStream(f)
        self._editor.insert(ts.readAll())
        del ts

        f.close()

    def _on_accepted(self):
        """ Invoked when the user accepts the edited text. """

        self.editDone.emit(True, self._editor.text())
        self._dialog.deleteLater()

    def _on_rejected(self):
        """ Invoked when the user rejects the edited text. """

        self.editDone.emit(False, '')
        self._dialog.deleteLater()



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QMessageBox


def question(title, text, detail=None, parent=None):
    """ Display a question to the user and return True if the answer was in the
    affirmative.
    """

    return QMessageBox.question(parent, title, text) is QMessageBox.StandardButton.Yes



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QCheckBox, QComboBox

from ...helpers import BaseDialog
from ...shell import EventType

from ..helpers import tagged_items

from .helpers import init_platform_selector


class DeletePlatformDialog(BaseDialog):
    """ This class implements the dialog for deleting a platform. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        self._platform = QComboBox()
        layout.addWidget(self._platform)

        self._discard = QCheckBox(
                "Discard platform-specific parts of the project?")
        layout.addWidget(self._discard)

    def set_fields(self):
        """ Set the dialog's fields from the project. """

        init_platform_selector(self._platform, self.model)

    def get_fields(self):
        """ Update the project from the dialog's fields. """

        project = self.model

        platform = self._platform.currentText()
        discard = self._discard.isChecked()

        # Delete from each API item it appears.
        remove_items = []

        for api_item, container in tagged_items(project):
            # Ignore items that aren't tagged with a platform.
            if len(api_item.platforms) == 0:
                continue

            remove_platforms = []

            for p in api_item.platforms:
                if p[0] == '!':
                    if p[1:] == platform:
                        # Just remove it if it is not the only one or if we are
                        # discarding if enabled (and the platform is inverted).
                        if len(api_item.platforms) > 1 or discard:
                            remove_platforms.append(f)
                        else:
                            remove_items.append((api_item, container))
                            break

                elif p == platform:
                    # Just remove it if it is not the only one or if we are
                    # discarding if enabled.
                    if len(api_item.platforms) > 1 or not discard:
                        remove_platforms.append(f)
                    else:
                        remove_items.append((api_item, container))
                        break
            else:
                # Note that we deal with a platform appearing multiple times,
                # even though that is probably a user bug.
                for p in remove_platforms:
                    api_item.platforms.remove(p)

        for api_item, container_item in remove_items:
            container_item.content.remove(api_item)
            self.shell.notify(EventType.CONTAINER_API_DELETE,
                    (container_item, api_item))

        # Delete from the project's list.
        project.platforms.remove(platform)

        self.shell.notify(EventType.PLATFORM_DELETE, platform)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QComboBox

from ...helpers import BaseDialog
from ...shell import EventType

from .helpers import init_module_selector


class DeleteModuleDialog(BaseDialog):
    """ This class implements the dialog for deleting a module. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        self._module = QComboBox()
        layout.addWidget(self._module)

    def set_fields(self):
        """ Set the dialog's fields from the project. """

        init_module_selector(self._module, self.model)

    def get_fields(self):
        """ Update the project from the dialog's fields. """

        project = self.model

        module_name = self._module.currentText()

        # Delete from the project's list.
        for module in project.modules:
            if module.name == module_name:
                project.modules.remove(module)
                break
        else:
            project.externalmodules.remove(module)

        self.shell.notify(EventType.MODULE_DELETE, module)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass


@dataclass
class Docstring:
    """ This class is a mixin for APIs that can have a docstring. """

    # The optional doc string.
    docstring: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


# In metasip prior to v2.12 the project format version was a single integer and
# some versions were known to require input from the user in order to update
# the project to that version.  For v2.12 and later the format version is a
# 2-tuple of major and minor version.  A change in the major version means user
# input is needed and a project cannot be used until the project is updated.  A
# change in the minor version means the project can be viewed and will only be
# updated to the latest minor version if the project is saved.
#
# A minimum project version is specified so that any version prior to this is
# not supported by metasip v2.12 and later.  An earlier version of metasip must
# first be used to update the project to that version.  Old versions were never
# seen "in the wild" so there is no need to carry the baggage of supporting
# them in current code.
#
# The project format version is stored in the project differently according to
# the major version of the format.  For major version 0 it is stored in the
# 'version' attribute.  For other major versions it is stored in the
# 'majorversion' and 'minorversion' attributes.


# Project format version history:
#
#  0.17 Implemented by metasip v2.13.
#       - Added 'comments' to the 'Class', 'Constructor', 'Destructor', 'Enum',
#         'EnumValue', 'Function', 'ManualCode', 'Method', 'Namespace',
#         'OpaqueClass', 'OperatorCast', 'OperatorFunction', 'OperatorMethod',
#         'Typedef' and 'Variable' elements.
#       - Added 'docstring' to the 'Typedef' element.
#       - Added 'keywordarguments' to the 'Module' element.
#       - Removed 'outputdirsuffix' from the 'Module' element.
#       - Added 'platforms' to the 'HeaderDirectory' element.
#       - Removed 'filefilter', 'inputdirsuffix' and 'parserargs' from the
#         'HeaderDirectory' element.
#
#  0.16 Implemented by metasip v2.11.
#       - Added 'pyssizetclean' to the 'Module' element.
#
#  0.15 Implemented by metasip v2.8.
#       - Removed 'version' from the 'Module' element.
#       - Added the 'removed' value of the 'status' attribute.


# The oldest supported project format.
MinimumProjectVersion = (0, 15)

# The latest supported project format.
ProjectVersion = (0, 17)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


class UserException(Exception):
    """ An exception capturing user friendly information. """

    def __init__(self, text, *, detail=None):
        """ Initialise the exception with its user friendly text and the
        optional detail.
        """

        super().__init__()

        self.text = text
        self.detail = detail



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from .base_dialog import BaseDialog
from .project_ui import ProjectUi
from .question import question
from .warning import warning



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


def adapt(model, target_type=None):
    """ Return an adapter for a model adapted to its type or super-type. """

    from .adapter_map import ADAPTER_MAP

    if target_type is None:
        target_type = type(model)

    assert isinstance(model, target_type)

    adapter_factory = ADAPTER_MAP[target_type]

    return adapter_factory(model)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QComboBox, QFormLayout, QLineEdit

from ...helpers import BaseDialog
from ...shell import EventType

from ..helpers import tagged_items

from .helpers import init_module_selector, validate_module_name


class RenameModuleDialog(BaseDialog):
    """ This class implements the dialog for renaming a module. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        self._module = QComboBox()
        layout.addWidget(self._module)

        form = QFormLayout()
        layout.addLayout(form)

        self._new_name = QLineEdit()
        form.addRow("New name", self._new_name)

    def set_fields(self):
        """ Set the dialog's fields from the project. """

        init_module_selector(self._module, self.model)

    def get_fields(self):
        """ Update the project from the dialog's fields. """

        project = self.model

        old_name = self._module.currentText()

        new_name = self._new_name.text().strip()
        if not validate_module_name(new_name, project, self):
            return False

        # Rename in the project's list.
        for module in project.modules:
            if module.name == old_name:
                module.name = new_name
                self.shell.notify(EventType.MODULE_RENAME, module)
                break
        else:
            project.externalmodules[project.externalmodules.index(old_name)] = new_name

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from .base_adapter import AttributeType, BaseAdapter


class DocstringAdapter(BaseAdapter):
    """ This is the Docstring adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'docstring':    AttributeType.LITERAL,
    }

    def generate_sip_directives(self, output):
        """ Write any directives to a .sip file. """

        output.write_code_directive('%Docstring', self.model.docstring,
                indent=False)

    def save_subelements(self, output):
        """ Save the XML subelements. """

        self.save_literal('docstring', output)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..helpers import validate_identifier, validation_error


def init_module_selector(selector, project):
    """ Initialise a module selector. """

    module_names = [module.name for module in project.modules]
    module_names.extend(project.externalmodules)

    selector.clear()
    selector.addItems(sorted(module_names))


def validate_module_name(module_name, project, dialog):
    """ Validate a module name and return True if it is valid. """

    message = validate_identifier(module_name, "module")
    if message is None:
        for module in project.modules:
            if module.name == module_name:
                message = "A module has already been defined with the same name."
                break
        else:
            if module_name in project.externalmodules:
                message = "An external module has already been defined with the same name."

    if message is not None:
        validation_error(message, dialog)
        return False

    return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QApplication, QProgressDialog

from ...project_io import AbstractProjectUi

from .warning import warning


class ProjectUi(AbstractProjectUi):
    """ This class encapsulates the UI-related methods supporting the loading
    of a project.
    """

    def error_creating_file(self, title, text, detail):
        """ Called when there was an error when creating a file. """

        warning(title, text, detail=detail)

    def load_starting(self, project, nr_steps):
        """ Called to initialise the UI prior to loading the project that will
        take a specific number of steps.
        """

        self._progress = QProgressDialog("Reading the project...", None, 0,
                nr_steps)
        self._progress.setWindowTitle(project.name)
        self._progress.setValue(0)

    def load_step(self):
        """ Called to update the UI once the next step of loading the project
        has been completed.
        """

        self._progress.setValue(self._progress.value() + 1)
        QApplication.processEvents()

    def update_project_format(self, root_element, from_version, to_version):
        """ Called to update the project from it's current major version before
        it is parsed.  Return True if the user didn't cancel.
        """

        # At the moment there is only one major version number.
        return True

    def warn_minor_version_update(self, from_version, to_version):
        """ Called to warn the user that the project will be updated to the
        current minor version if saved.
        """

        from_s = f'{from_version[0]}.{from_version[1]}'
        to_s = f'{to_version[0]}.{to_version[1]}'

        warning("Update project format",
                f"The format of this project is v{from_s}. It will be automatically updated to v{to_s} if you save it.")



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass, field

from .header_file_version import HeaderFileVersion


@dataclass
class HeaderFile:
    """ This class implements a C/C++ .h file. """

    # This is set if the header file is being ignored, i.e. it will never be
    # assigned to a module.
    ignored: bool = False

    # The name of the optional module that the header file is assigned to.
    module: str = ''

    # The basename of the header file.
    name: str = ''

    # The individual versions of the header file.  The list will be empty if
    # it is being ignored.  If no versions have been defined the the list will
    # have one element.  Note that these are unordered.
    versions: list[HeaderFileVersion] = field(default_factory=list)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from abc import ABC, abstractmethod

from PyQt6.QtWidgets import (QDialog, QDialogButtonBox, QSizePolicy,
        QSpacerItem, QVBoxLayout)


class BaseDialog(ABC):
    """ A base class for dialogs that update a model. """

    def __init__(self, model, title, shell):
        """ Initialise the dialog. """

        self.model = model
        self.shell = shell

        self.widget = QDialog(shell.widget)

        self.widget.setWindowTitle(title)

        layout = QVBoxLayout()
        self.widget.setLayout(layout)

        self.populate(layout)

        spacer = QSpacerItem(1, 1, QSizePolicy.Policy.Minimum,
                QSizePolicy.Policy.Expanding)
        layout.addItem(spacer)

        button_box = QDialogButtonBox(
                QDialogButtonBox.StandardButton.Cancel |
                QDialogButtonBox.StandardButton.Ok)
        button_box.accepted.connect(self._handle_accept)
        button_box.rejected.connect(self.widget.reject)
        layout.addWidget(button_box)

    @abstractmethod
    def populate(self, layout):
        """ Reimplemented by a sub-class to populate the dialog's layout. """

        ...

    def update(self):
        """ Return True if the model was updated or False if the dialog was
        cancelled.
        """

        self.set_fields()

        return self.widget.exec() == int(QDialog.DialogCode.Accepted)

    def set_fields(self):
        """ Normally reimplemented by a sub-class to set the dialog's fields
        from the model.
        """

        # This default implementation does nothing.
        pass

    def get_fields(self):
        """ Reimplemented by a sub-class to update the model from the
        dialog's fields.  Returns True if the update was valid.
        """

        # This default implementation does nothing.
        return True

    def _handle_accept(self):
        """ Handle the Ok button. """

        # Only accept the dialog if the fields are valid. """
        if self.get_fields():
            self.widget.accept()




# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass


@dataclass
class Platform:
    """ This class implements a header directory and platform specific
    configuration.
    """

    # The suffix added to the input directory to create the glob pattern that
    # defines the input files.
    inputdirpattern: str = ''

    # The name of the platform.  This is used for display purposes.
    name: str = ''

    # The optional additional arguments passed to the external C++ parser.
    parserargs: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..exceptions import UserException
from ..models.adapters import adapt

from .indent_file import IndentFile


def save_project(project, ui):
    """ Save a project to its project file.  Return True if there was no error.
    """

    try:
        output = IndentFile.create(project.name, indent=2)
    except UserException as e:
        ui.error_creating_file("Save", e.text, e.detail)
        return False

    adapt(project).save(output)

    output.close()

    return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .code import Code
from .docstring import Docstring


@dataclass
class Typedef(Code, Docstring):
    """ This class implements a typedef. """

    # The name of the type.
    name: str = ''

    # The definition of the type.
    type: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QGridLayout, QLabel, QLineEdit, QPlainTextEdit

from ....helpers import BaseDialog
from ....shell import EventType


class ProjectPropertiesDialog(BaseDialog):
    """ This class implements the dialog for a project's properties. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        grid_layout = QGridLayout()
        layout.addLayout(grid_layout)

        grid_layout.addWidget(QLabel("Root package"), 0, 0)
        self._root_module = QLineEdit()
        grid_layout.addWidget(self._root_module, 0, 1, 1, 2)

        grid_layout.addWidget(QLabel("SIP file comments"), 1, 0,
                Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop)
        self._sip_comments = QPlainTextEdit()
        self._sip_comments.setLineWrapMode(QPlainTextEdit.LineWrapMode.NoWrap)
        grid_layout.addWidget(self._sip_comments, 1, 1, 1, 2)

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        self._root_module.setText(self.model.rootmodule)
        self._sip_comments.setPlainText(self.model.sipcomments)

    def get_fields(self):
        """ Update the API item from the dialog's fields. """

        self.model.rootmodule = self._root_module.text().strip()
        self.model.sipcomments = self._sip_comments.toPlainText().strip()

        self.shell.notify(EventType.PROJECT_ROOT_MODULE_RENAME)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..annos import Annos
from ..tagged import Tagged
from ..workflow import Workflow

from .adapt import adapt
from .base_adapter import BaseAdapter


class CodeAdapter(BaseAdapter):
    """ This is the Code adapter. """

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        adapt(self.model, Annos).load(element, project, ui)
        adapt(self.model, Tagged).load(element, project, ui)
        adapt(self.model, Workflow).load(element, project, ui)

    def save_attributes(self, output):
        """ Save the XML attributes. """

        # The order is to match older versions.
        adapt(self.model, Annos).save_attributes(output)
        adapt(self.model, Workflow).save_attributes(output)
        adapt(self.model, Tagged).save_attributes(output)

    def save_subelements(self, output):
        """ Save the XML subelements. """

        adapt(self.model, Annos).save_subelements(output)
        adapt(self.model, Tagged).save_subelements(output)
        adapt(self.model, Workflow).save_subelements(output)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


import os
import sys

from ...helpers import get_platform_name, get_supported_platforms

from ..header_file import HeaderFile
from ..platform import Platform

from .adapt import adapt
from .base_adapter import AttributeType, BaseAdapter


class HeaderDirectoryAdapter(BaseAdapter):
    """ This is the HeaderDirectory adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'name':             AttributeType.STRING,
    }

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        super().load(element, project, ui)

        header_directory = self.model

        scan = element.get('scan')
        if scan is None:
            scan = []
        elif scan == '':
            scan = ['']
        else:
            scan = scan.split()

        header_directory.scan = scan

        # Platforms were introduced in project format v0.17.
        if project.version < (0, 17):
            # Create a Platform for legacy versions.
            filefilter = element.get('filefilter')
            inputdirsuffix = element.get('inputdirsuffix')
            parserargs = element.get('parserargs')

            inputdirpattern = os.path.join(inputdirsuffix, filefilter)

            # The C++ standard was hardcoded (so we enforce it here) but should
            # be configurable.
            parserargs = '-std=c++17 ' + parserargs

            header_directory.platforms.append(
                    Platform(name=get_platform_name(),
                            inputdirpattern=inputdirpattern,
                            parserargs=parserargs))

        for subelement in element:
            if subelement.tag == 'HeaderFile':
                header_file = HeaderFile()
                adapt(header_file).load(subelement, project, ui)
                header_directory.content.append(header_file)
            elif subelement.tag == 'Platform':
                platform = Platform()
                adapt(platform).load(subelement, project, ui)
                header_directory.platforms.append(platform)

        # Supply defaults for any missing supported platforms but don't mark
        # the project as dirty so they are only saved if something else has
        # changed.
        platform_names = [p.name for p in header_directory.platforms]

        for supported in get_supported_platforms():
            if supported not in platform_names:
                # On macOS we assume the library being wrapped is a framework.
                if supported == 'macOS':
                    inputdirpattern = header_directory.name + '.framework/Header/*.h'
                    parserargs = '-F .'
                else:
                    inputdirpattern = '*.h'
                    parserargs = ''

                header_directory.platforms.append(
                        Platform(name=supported,
                                inputdirpattern=inputdirpattern,
                                parserargs=parserargs))

    def save(self, output):
        """ Save the model to an output file. """

        header_directory = self.model

        output.write(f'<HeaderDirectory name="{header_directory.name}"')

        if header_directory.scan:
            if header_directory.scan[0] == '':
                scan = ''
            else:
                scan = ' '.join(header_directory.scan)

            self.save_attribute('scan', scan, output)

        output.write('>\n')

        output += 1

        for platform in header_directory.platforms:
            adapt(platform).save(output)

        for header_file in header_directory.content:
            adapt(header_file).save(output)

        output -= 1
        output.write('</HeaderDirectory>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


def version_range(version_range):
    """ Return a version range as a string. """

    if version_range.startversion == '':
        if version_range.endversion == '':
            # This should never happen.
            return ''

        return '- ' + version_range.endversion

    if version_range.endversion == '':
        return version_range.startversion + ' -'

    return version_range.startversion + ' - ' + version_range.endversion



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..annos import Annos
from ..callable import Callable

from .adapt import adapt
from .base_adapter import BaseApiAdapter


class OperatorFunctionAdapter(BaseApiAdapter):
    """ This is the OperatorFunction adapter. """

    def __eq__(self, other):
        """ Compare for C/C++ equality. """

        arg = self.model
        other_arg = other.model

        if type(arg) is not type(other_arg):
            return False

        if adapt(function, Callable) != adapt(other_function, Callable):
            return False

        return True

    def as_str(self):
        """ Return the standard string representation. """

        function = self.model

        callable_adapter = adapt(function, Callable)

        s = callable_adapter.return_type_as_str(allow_py=True) + 'operator' + function.name

        if function.pyargs != '':
            s += function.pyargs
        else:
            args = ', '.join([adapt(arg).as_py_str() for arg in function.args])
            s += '(' + args + ')'

        s += adapt(function, Annos).as_str()

        if callable_adapter.has_different_signatures():
            return_type = callable_adapter.return_type_as_str().strip()
            args = ', '.join([adapt(arg).as_str() for arg in function.args])
            s += f' [{return_type} ({args})]'

        return s

    def generate_sip(self, sip_file, output):
        """ Generate the .sip file content. """

        function = self.model

        nr_ends = self.version_start(output)

        output.write(self.as_str())
        output.write(';\n')
        adapt(function, Callable).generate_sip_directives(output)

        self.version_end(nr_ends, output)

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        adapt(self.model, Callable).load(element, project, ui)

    def save(self, output):
        """ Save the model to an output file. """

        function = self.model

        output.write('<OperatorFunction')
        adapt(function, Callable).save_attributes(output)
        output.write('>\n')

        output += 1
        adapt(function, Callable).save_subelements(output)
        output -= 1

        output.write('</OperatorFunction>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .access import Access
from .code import Code


@dataclass
class Destructor(Code, Access):
    """ This class implements a C++ destructor. """

    # The optional %MethodCode.
    methcode: str = ''

    # The name of the destructor (ie. the class).
    name: str = ''

    # The optional %VirtualCatcherCode.
    virtcode: str = ''

    # This is set if the destructor is virtual.
    virtual: bool = False



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..annos import Annos
from ..callable import Callable
from ..docstring import Docstring

from .adapt import adapt
from .base_adapter import BaseApiAdapter


class FunctionAdapter(BaseApiAdapter):
    """ This is the Function adapter. """

    def __eq__(self, other):
        """ Compare for C/C++ equality. """

        function = self.model
        other_function = other.model

        if type(function) is not type(other_function):
            return False

        if adapt(function, Callable) != adapt(other_function, Callable):
            return False

        return True

    def as_str(self):
        """ Return the standard string representation. """

        return adapt(self.model, Callable).as_str()

    def generate_sip(self, sip_file, output):
        """ Generate the .sip file content. """

        function = self.model

        nr_ends = self.version_start(output)

        # Note that we don't use CallableAdapter's implementation because we
        # dont want any C/C++ signature.

        s = adapt(function, Callable).return_type_as_str(allow_py=True) + function.name

        if function.pyargs != '':
            s += function.pyargs
        else:
            args = ', '.join([adapt(arg).as_py_str() for arg in function.args])
            s += '(' + args + ')'

        s += adapt(function, Annos).as_str()
        output.write(s)
        output.write(';\n')

        adapt(function, Docstring).generate_sip_directives(output)
        adapt(function, Callable).generate_sip_directives(output)

        self.version_end(nr_ends, output)

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        adapt(self.model, Callable).load(element, project, ui)
        adapt(self.model, Docstring).load(element, project, ui)

    def save(self, output):
        """ Save the model to an output file. """

        function = self.model

        output.write('<Function')
        adapt(function, Callable).save_attributes(output)
        adapt(function, Docstring).save_attributes(output)
        output.write('>\n')

        output += 1
        # The order is to match older versions.
        adapt(function, Docstring).save_subelements(output)
        adapt(function, Callable).save_subelements(output)
        output -= 1

        output.write('</Function>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .access import Access
from .callable import Callable


@dataclass
class OperatorCast(Callable, Access):
    """ This class implements a cast operator. """

    # This is set if the operator is const.
    const: bool = False



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..annos import Annos
from ..argument import Argument
from ..code import Code

from .adapt import adapt
from .base_adapter import AttributeType, BaseAdapter


class CallableAdapter(BaseAdapter):
    """ This is the Callable adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'methcode': AttributeType.LITERAL,
        'name':     AttributeType.STRING,
        'pyargs':   AttributeType.STRING,
        'pytype':   AttributeType.STRING,
        'rtype':    AttributeType.STRING,
    }

    def __eq__(self, other):
        """ Compare for C/C++ equality. """

        callable = self.model
        other_callable = other.model

        # Note that we have already checked the types.

        if callable.name != other_callable.name:
            return False

        if self.expand_type(callable.rtype) != self.expand_type(other_callable.rtype):
            return False

        if len(callable.args) != len(other_callable.args):
            return False

        for arg, other_arg in zip(callable.args, other_callable.args):
            if adapt(arg) != adapt(other_arg):
                return False

        return True

    def as_str(self):
        """ Return the standard string representation. """

        callable = self.model

        # This is the Python signature.
        s = self.return_type_as_str(allow_py=True) + callable.name

        if callable.pyargs != '':
            s += callable.pyargs
        else:
            args = ', '.join([adapt(arg).as_py_str() for arg in callable.args])
            s += '(' + args + ')'

        s += adapt(callable, Annos).as_str()

        # We include a separate C++ signature if it is different to the Python
        # signature.  This is so we always hint to the user that something has
        # been manually changed.
        if self.has_different_signatures():
            return_type = self.return_type_as_str().strip()
            if return_type != '':
                return_type += ' '

            args = ', '.join([adapt(arg).as_str() for arg in callable.args])
            s += f' [{return_type}({args})]'

        return s

    def has_different_signatures(self):
        """ Returns True if the Python and C/C++ signatures are different. """

        callable = self.model

        if callable.pytype != '' or callable.pyargs != '':
            return True

        for arg in callable.args:
            if arg.pytype != '':
                return True

        return False

    def generate_sip_directives(self, output):
        """ Write any directives to a .sip file. """

        output.write_code_directive('%MethodCode', self.model.methcode)

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        super().load(element, project, ui)

        adapt(self.model, Code).load(element, project, ui)

        for subelement in element:
            if subelement.tag == 'Argument':
                arg = Argument()
                adapt(arg).load(subelement, project, ui)
                self.model.args.append(arg)

    def return_type_as_str(self, allow_py=False):
        """ Return the return type as a string. """

        callable = self.model

        if callable.pytype != '' and allow_py:
            s = callable.pytype
        elif callable.rtype != '':
            s = self.expand_type(callable.rtype)
        else:
            return ''

        if s[-1] not in '&*':
            s += ' '

        return s

    def save_attributes(self, output):
        """ Save the XML attributes. """

        callable = self.model

        adapt(callable, Code).save_attributes(output)
        self.save_attribute('name', callable.name, output)
        self.save_str('rtype', output)
        self.save_str('pytype', output)
        self.save_str('pyargs', output)

    def save_subelements(self, output):
        """ Save the XML subelements. """

        adapt(self.model, Code).save_subelements(output)

        for argument in self.model.args:
            adapt(argument).save(output)

        self.save_literal('methcode', output)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QAbstractSlider, QApplication, QPlainTextEdit

from ..shell import EventType
from ..shell_tool import ShellTool, ShellToolLocation


class LoggerTool(ShellTool):
    """ This class implements a tool for displaying the log. """

    def __init__(self, shell):
        """ Initialise the tool. """

        super().__init__(shell)

        self._log_widget = QPlainTextEdit(readOnly=True)

    def event(self, event_type, event_arg):
        """ Reimplemented to handle project-specific events. """

        if event_type is EventType.LOG_MESSAGE:
            self._log_message(event_arg)

    @property
    def location(self):
        """ Get the location of the tool in the shell. """

        return ShellToolLocation.BOTTOM

    @property
    def name(self):
        """ Get the tool's name. """

        return 'metasip.tool.logger'

    @property
    def title(self):
        """ Get the tool's title. """

        return "Log"

    @property
    def widget(self):
        """ Get the tool's widget. """

        return self._log_widget

    def _log_message(self, message):
        """ Log a message. """

        # Add the new message
        self._log_widget.appendPlainText(message)

        # Make sure the new message is visible.
        self._log_widget.verticalScrollBar().triggerAction(
                QAbstractSlider.SliderAction.SliderToMaximum)

        # Update the screen so that individual messages appear as soon as they
        # are logged.
        QApplication.processEvents()



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


import glob
import hashlib
import os

from PyQt6.QtWidgets import (QApplication, QCheckBox, QComboBox, QFileDialog,
        QFormLayout, QGridLayout, QGroupBox, QHBoxLayout, QInputDialog, QLabel,
        QLineEdit, QMessageBox, QPushButton, QStyle, QTabWidget, QToolButton,
        QVBoxLayout, QWidget)

from .....helpers import (get_platform_name, get_supported_platforms,
        header_directory_platform, VersionMap)
from .....models import (Callable, CodeContainer, Constructor, Enum,
        HeaderDirectory, HeaderFile, HeaderFileVersion, ManualCode, SipFile,
        VersionRange)
from .....models.adapters import adapt

from ....helpers import warning
from ....shell import EventType


class ControlWidget(QWidget):
    """ This class is a widget that implements the control part of a scanner's
    GUI.
    """

    def __init__(self, tool):
        """ Initialise the widget. """

        super().__init__()

        self._tool = tool
        self._header_directory = None
        self._header_file = None

        layout = QVBoxLayout()
        self.setLayout(layout)

        form = QFormLayout()
        layout.addLayout(form)

        self._working_version = QComboBox(
                currentTextChanged=self._handle_working_version)
        form.addRow("Working version", self._working_version)

        h_box = QHBoxLayout()
        form.addRow("Source directory", h_box)

        self._source_directory = QLineEdit(
                textChanged=self._configure_parse_button)
        h_box.addWidget(self._source_directory)

        button = QToolButton(
                icon=QApplication.style().standardIcon(
                        QStyle.StandardPixmap.SP_DirIcon),
                clicked=self._handle_browse_source_directory)
        h_box.addWidget(button)

        group_box = QGroupBox("Header Directory")
        layout.addWidget(group_box)

        v_box = QVBoxLayout()
        group_box.setLayout(v_box)

        self._header_directory_form = QFormLayout()
        v_box.addLayout(self._header_directory_form)

        self._header_directory_name = QLabel()
        self._header_directory_form.addRow("Name", self._header_directory_name)

        self._platforms = QTabWidget()
        v_box.addWidget(self._platforms)

        current_platform_name = get_platform_name()
        current_platform_page = None

        for platform_name in get_supported_platforms():
            page = QWidget()

            form = QFormLayout()
            page.setLayout(form)

            form.addRow("Header file pattern",
                    QLineEdit(objectName='inputdirpattern'))
            form.addRow("Parser arguments", QLineEdit(objectName='parserargs'))

            self._platforms.addTab(page, platform_name)

            if platform_name == current_platform_name:
                current_platform_page = page

        self._platforms.setCurrentWidget(current_platform_page)

        self._showing_ignored = QCheckBox("Show ignored header files?",
                stateChanged=self._handle_showing_ignored)
        v_box.addWidget(self._showing_ignored)

        grid = QGridLayout()
        v_box.addLayout(grid)

        self._scan_button = QPushButton("Scan",
                clicked=self._handle_scan_header_directory)
        grid.addWidget(self._scan_button, 0, 0)

        self._update_directory_button = QPushButton("Update",
                clicked=self._handle_update_header_directory_properties)
        grid.addWidget(self._update_directory_button, 0, 1)

        self._new_button = QPushButton("New...",
                clicked=self._handle_new_header_directory)
        grid.addWidget(self._new_button, 1, 0)

        self._delete_button = QPushButton("Delete...",
                clicked=self._handle_delete_header_directory)
        grid.addWidget(self._delete_button, 1, 1)

        group_box = QGroupBox("Header File")
        layout.addWidget(group_box)

        v_box = QVBoxLayout()
        group_box.setLayout(v_box)

        self._header_file_form = QFormLayout()
        v_box.addLayout(self._header_file_form)

        self._header_file_name = QLabel()
        self._header_file_form.addRow("Name", self._header_file_name)

        self._ignored = QCheckBox(
                stateChanged=self._handle_header_file_ignored)
        self._header_file_form.addRow("Ignored?", self._ignored)

        self._module = QComboBox()
        self._header_file_form.addRow("Module", self._module)

        grid = QGridLayout()
        v_box.addLayout(grid)

        self._parse_button = QPushButton("Parse",
                clicked=self._handle_parse_header_file)
        grid.addWidget(self._parse_button, 0, 0)

        self._update_file_button = QPushButton("Update",
                clicked=self._handle_update_header_file_properties)
        grid.addWidget(self._update_file_button, 0, 1)

        button = QPushButton("Reset workflow",
                clicked=self._handle_reset_workflow)
        layout.addWidget(button)

        layout.addStretch()

    def module_add_delete(self):
        """ Handle a change in the list of modules. """

        if self._header_file is not None:
            self._set_module_selector(self._header_file.ignored)

    def module_rename(self, module):
        """ A module has been renamed. """

        for index in range(self._module.count()):
            if self._module.itemData(index) is module:
                self._module.setItemText(index, module.name)

    def restore_state(self, settings):
        """ Restore the widget's state. """

        state = settings.value('source_directory')
        if state is not None:
            self._source_directory.setText(state)

    def save_state(self, settings):
        """ Save the widget's state. """

        settings.setValue('source_directory', self._source_directory.text())

    def set_header_file(self, header_file, header_directory, showing_ignored):
        """ Set the current header file. """

        self._header_file = header_file
        self._header_directory = header_directory

        if header_directory is None:
            self._header_directory_name.setText(None)

            for _, inputdirpattern, parserargs in self._get_platform_widgets():
                inputdirpattern.clear()
                parserargs.clear()

            self._showing_ignored.setChecked(False)

            enabled = False
        else:
            self._header_directory_name.setText(header_directory.name)

            for platform_name, inputdirpattern, parserargs in self._get_platform_widgets():
                platform = header_directory_platform(header_directory,
                        platform_name=platform_name)

                inputdirpattern.setText(platform.inputdirpattern)
                parserargs.setText(platform.parserargs)

            self._showing_ignored.setChecked(showing_ignored)

            enabled = True

        self._enable_layout(self._header_directory_form, enabled)
        self._showing_ignored.setEnabled(enabled)
        self._scan_button.setEnabled(enabled)
        self._update_directory_button.setEnabled(enabled)
        self._delete_button.setEnabled(enabled)

        if header_file is None:
            self._header_file_name.setText(None)
            self._module.clear()
            self._ignored.setChecked(False)

            enabled = False
        else:
            self._header_file_name.setText(header_file.name)
            self._ignored.setChecked(header_file.ignored)
            self._set_module_selector(header_file.ignored)

            enabled = True

        self._enable_layout(self._header_file_form, enabled)
        self._configure_parse_button()
        self._update_file_button.setEnabled(enabled)

    def set_project(self):
        """ Set the current project. """

        self.set_header_file(None, None, False)
        self._init_version_selector()

    def set_working_version(self, working_version):
        """ Set the current working version. """

        if working_version is None:
            self._working_version.clear()
            enabled = False
        else:
            self._working_version.setCurrentText(working_version)
            enabled = True

        self._working_version.setEnabled(enabled)

    def version_add_delete(self):
        """ A version has been added or deleted. """

        project = self._tool.shell.project

        if len(project.versions) == 0:
            working_version = None
        else:
            # Try and maintain the current working version.
            working_version = self._working_version.currentText()

            self._init_version_selector()

            if working_version not in project.versions:
                working_version = project.versions[-1]

        self._tool.set_working_version(working_version)

    def version_rename(self, old_name, new_name):
        """ A version has been renamed. """

        index = self._working_version.findText(old_name)
        self._working_version.setItemText(index, new_name)

    def _configure_parse_button(self):
        """ Configure the 'Parse' button. """

        # The button is enabled if there is a source directory and the selected
        # header file is assigned to a module.
        enabled = (self._source_directory.text() != '' and self._header_file is not None and self._header_file.module != '')

        self._parse_button.setEnabled(enabled)

    @staticmethod
    def _enable_layout(layout, enabled):
        """ Enable or disable all the items in a layout. """

        for item_nr in range(layout.count()):
            widget = layout.itemAt(item_nr).widget()
            if widget is not None:
                widget.setEnabled(enabled)

    def _handle_browse_source_directory(self):
        """ Handle the button to browse for a source directory. """

        source_directory = QFileDialog.getExistingDirectory(self,
                "Source Directory", self._source_directory.text())

        if source_directory:
            self._source_directory.setText(source_directory)

    def _handle_delete_header_directory(self):
        """ Handle the button to delete a header directory. """

        button = QMessageBox.question(self, "Delete Header Directory",
                "Are you sure you want to delete this header directory?")

        if button is QMessageBox.StandardButton.Yes:
            self._tool.shell.project.headers.remove(self._header_directory)
            self._tool.header_directory_removed(self._header_directory)
            self._tool.shell.dirty = True

    def _handle_header_file_ignored(self, state):
        """ Handle the checkbox to toggle an ignored header file. """

        self._set_module_selector(bool(state))

    def _handle_new_header_directory(self):
        """ Handle the button to add a header directory. """

        name, ok = QInputDialog.getText(self, "New Header Directory",
                "Descriptive name")

        if ok:
            working_version = self._working_version.currentText()
            header_directory = HeaderDirectory(name=name,
                    scan=[working_version])
            self._tool.shell.project.headers.append(header_directory)
            self._tool.header_directory_added(header_directory,
                    working_version)
            self._tool.shell.dirty = True

    def _handle_parse_header_file(self):
        """ Handle the button to parse a header file. """

        from ..cast_xml import CastXMLParser

        project = self._tool.shell.project
        header_directory = self._header_directory
        header_file = self._header_file
        source_directory = self._source_directory.text()

        parser = CastXMLParser()

        header_directory_name = os.path.dirname(
                header_directory_platform(header_directory).inputdirpattern)
        name = os.path.join(source_directory, header_directory_name,
                header_file.name)

        if not os.access(name, os.R_OK):
            warning("Parse", f"Unable to read '{name}'.", parent=self)
            return

        _, name, _ = self._read_header(name)

        parsed_header_file = parser.parse(project, source_directory,
                header_directory, header_file, name, self._tool.shell.log)

        if parsed_header_file is None:
            warning("Parse", parser.diagnostic, parent=self)
            return

        # Find the corresponding .sip file creating it if is a new header file.
        # FIXME: Assuming we ultimately want to be able to create a complete
        #        project without parsing .h files then we will need the ability
        #        (in the main editor) to manually create a SipFile instance.
        for module in project.modules:
            if module.name == header_file.module:
                for sip_file in module.content:
                    if sip_file.name == header_file.name:
                        break
                else:
                    sip_file = SipFile(name=header_file.name)
                    module.content.append(sip_file)
                    self._tool.shell.notify(EventType.CONTAINER_API_ADD,
                            module, sip_file)

                self._merge_code(sip_file, parsed_header_file)
                break

        # The file version no longer needs parsing.
        working_version = self._working_version.currentText()

        for header_file_version in header_file.versions:
            if header_file_version.version == working_version:
                header_file_version.parse = False
                self._tool.header_file_status(header_file)
                break

        self._tool.shell.dirty = True

    def _handle_reset_workflow(self):
        """ Handle the button to reset the workflow. """

        working_version = self._working_version.currentText()

        for header_directory in self._tool.shell.project.headers:
            if working_version == '':
                header_directory.scan = ['']
            elif working_version not in header_directory.scan:
                header_directory.scan.append(working_version)

            self._tool.header_directory_status(header_directory)

        self._tool.shell.dirty = True

    def _handle_scan_header_directory(self):
        """ Handle the button to scan a header directory. """

        shell = self._tool.shell
        project = shell.project
        header_directory = self._header_directory
        platform = header_directory_platform(header_directory)

        source_pattern = os.path.abspath(self._source_directory.text())
        source_pattern = os.path.join(source_directory,
                platform.inputdirpattern)

        header_directory_path = os.path.dirname(source_pattern)
        shell.log(f"Scanning header directory '{header_directory_path}'")

        # Save the files that were in the directory.
        saved = list(header_directory.content)

        for header_path in glob.iglob(source_pattern):
            if not os.path.isfile(header_path):
                continue

            if os.access(header_path, os.R_OK):
                header_file = self._scan_header_file(header_path)

                for saved_header_file in saved:
                    if saved_header_file is header_file:
                        saved.remove(saved_header_file)
                        break
                else:
                    # It's a new header file.
                    header_directory.content.append(header_file)
                    shell.dirty = True

                shell.log(f"Scanned '{header_path}'")
            else:
                shell.log(f"Skipping unreadable header file '{header_path}'")

        # Anything left in the saved list has gone missing or was already
        # missing.
        working_version = self._working_version.currentText()

        for header_file in saved:
            for header_file_version in header_file.versions:
                if header_file_version.version == working_version:
                    header_file.versions.remove(header_file_version)

                    # If there is only one version left then remove the file
                    # itself.
                    if len(header_file.versions) == 0:
                        header_directory.content.remove(header_file)
                        self._remove_from_module(header_file)
                    else:
                        # FIXME: Go through the corresponding SipFile and make
                        # sure that all top-level APIs have an upper version
                        # set.
                        pass

                    shell.log(f"'{header_file.name}' is no longer in the header directory")

                    self._tool.header_file_removed(header_file)
                    shell.dirty = True
                    break

        # This version no longer needs scanning.
        if working_version in header_directory.scan:
            header_directory.scan.remove(working_version)
            self._tool.header_directory_status(header_directory)
            shell.dirty = True

    def _handle_showing_ignored(self, state):
        """ Handle the checkbox to toggle ignored header files. """

        self._tool.set_header_files_visibility(self._header_directory,
                bool(state))

    def _handle_working_version(self, new_working_version):
        """ Handle the user changing the working version. """

        self._tool.set_working_version(new_working_version)

    def _handle_update_header_directory_properties(self):
        """ Handle the button to update a header directory's properties. """

        for platform_name, inputdirpattern, parserargs in self._get_platform_widgets():
            platform = header_directory_platform(self._header_directory,
                    platform_name=platform_name)

            platform.inputdirpattern = inputdirpattern.text()
            platform.parserargs = parserargs.text()

        self._tool.shell.dirty = True

    def _handle_update_header_file_properties(self):
        """ Handle the button to update a header file's properties. """

        header_file = self._header_file

        header_file.ignored = self._ignored.isChecked()

        # Make sure the rest is consistent.
        if header_file.ignored:
            header_file.module = ''
            header_file.versions = []
        else:
            header_file.module = self._module.currentText()

            # We may just have un-ignored the header file.
            if len(header_file.versions) == 0:
                header_file.versions.append(
                        HeaderFileVersion(parse=True,
                                version=self._working_version.currentText()))

        self._configure_parse_button()
        self._tool.header_file_status(header_file)
        self._tool.shell.dirty = True

    def _init_version_selector(self):
        """ Initialise the version selector. """

        blocked = self._working_version.blockSignals(True)
        self._working_version.clear()
        self._working_version.addItems(self._tool.shell.project.versions)
        self._working_version.blockSignals(blocked)

    def _merge_code(self, dst_code, src_code):
        """ Merge source code into destination code. """

        working_version = self._working_version.currentText()

        # Go though each existing code API.
        for dst_api in list(dst_code.content):
            # Manual code is always retained.
            if isinstance(dst_api, ManualCode):
                continue

            # Go through each potentially new code API.
            for src_api in list(src_code):
                if adapt(dst_api) == adapt(src_api):
                    # Make sure the versions include the working version.
                    if working_version != '':
                        self._add_working_version(dst_api)

                    # Discard the new code API.
                    src_code.remove(src_api)

                    # Merge any child code.
                    if isinstance(dst_api, (CodeContainer, Enum)):
                        self._merge_code(dst_api, src_api.content)

                    break
            else:
                # The existing one doesn't exist in the working version.
                if working_version == '':
                    # If it is ignored then forget about it because there are
                    # no other versions that might refer to it.
                    if dst_api.status == 'ignored':
                        dst_code.content.remove(dst_api)
                        self._tool.shell.notify(
                                EventType.CONTAINER_API_DELETE,
                                (dst_code, dst_api))
                    else:
                        dst_api.status = 'removed'
                        self._tool.shell.notify(EventType.API_STATUS, dst_api)
                else:
                    version_status = self._remove_working_version(dst_api)
                    if version_status == 'no_longer_working':
                        # It's removal needs checking.
                        if dst_api.status == '':
                            dst_api.status = 'unknown'
                            self._tool.shell.notify(EventType.API_STATUS,
                                    dst_api)
                    elif version_status == 'no_longer_any':
                        # Forget about it because there are no other versions
                        # that refer to it.
                        dst_code.content.remove(dst_api)
                        self._tool.shell.notify(
                                EventType.CONTAINER_API_DELETE,
                                (dst_code, dst_api))

        # Anything left in the source code is new.

        if working_version == '':
            startversion = endversion = ''
        else:
            versions = self._tool.shell.project.versions
            working_idx = versions.index(working_version)

            # If the working version is the first then assume that the new API
            # will appear in earlier versions, otherwise it is restricted to
            # this version.
            startversion = '' if working_idx == 0 else working_version

            # If the working version is the latest then assume that the new API
            # will appear in later versions, otherwise it is restricted to this
            # version.
            try:
                endversion = versions[working_idx + 1]
            except IndexError:
                endversion = ''

        for src_api in src_code:
            if startversion != '' or endversion != '':
                src_api.versions.append(
                        VersionRange(startversion=startversion,
                                endversion=endversion))
                self._tool.shell.notify(EventType.API_VERSIONS, src_api)

            # Try and place the new API with any similar one.
            pos = -1
            for idx, code in enumerate(dst_code.content):
                if type(code) is not type(src_api):
                    continue

                if isinstance(src_api, Constructor):
                    pos = idx
                    break

                if isinstance(src_api, Callable) and code.name == src_api.name:
                    pos = idx
                    break

            if pos >= 0:
                dst_code.content.insert(pos, src_api)
            else:
                dst_code.content.append(src_api)

            self._tool.shell.notify(EventType.CONTAINER_API_ADD,
                    (dst_code, src_api))

    def _add_working_version(self, api):
        """ Add the working version to an API's version ranges. """

        # There is only something to do if the API is currently versioned.
        if len(api.versions) != 0:
            project = self._tool.shell.project

            # Add the working version.
            vmap = VersionMap(project, api.versions)
            vmap[self._working_version.currentText()] = True
            api.versions = vmap.as_version_ranges()

            self._tool.shell.notify(EventType.API_VERSIONS, api)

    def _remove_working_version(self, api):
        """ Remove the working version from an API's version ranges.  Returns
        'wasnt_working' if the API wasn't in the working version,
        'no_longer_working' if the API is no longer in the working version and
        'no_longer_any' if the API is no longer in any version.
        """

        project = self._tool.shell.project

        # Construct the existing list of version ranges to a version map.
        vmap = VersionMap(project, api.versions)

        # Update the version map appropriately using the working version.
        # First take a shortcut to see if anything has changed.
        working_version = self._working_version.currentText()

        if not vmap[working_version]:
            return 'wasnt_working'

        vmap[working_version] = False

        # Convert the version map back to a list of version ranges.
        versions = vmap.as_version_ranges()

        if versions is None:
            return 'no_longer_any'

        api.versions = versions
        self._tool.shell.notify(EventType.API_VERSIONS, api)

        return 'no_longer_working'

    @classmethod
    def _read_header(cls, name):
        """ Read the contents of a header file.  Handle the special case of the
        file just being a #include redirect to another header file.
        """

        contents, actual_name, encoding = cls._read_single_header(name)
        while actual_name != name:
            name = actual_name
            contents, actual_name, encoding = cls._read_single_header(name)

        return contents, actual_name, encoding

    @staticmethod
    def _read_single_header(name):
        """ Read the contents of a single header file. """

        with open(name, 'r') as f:
            contents = f.read()
            encoding = f.encoding

        lines = contents.strip().split('\n')
        if len(lines) == 1:
            words = lines[0].split()
            if len(words) == 2 and words[0] == '#include':
                include_name = words[1]
                if include_name.startswith('".') and include_name.endswith('"'):
                    name = os.path.dirname(name) + '/' + include_name[1:-1]

        return contents, name, encoding

    def _remove_from_module(self, header_file):
        """ Handle the removal of a header file from the project. """

        # Find the corresponding .sip file.
        for mod in self.current_project.modules:
            if mod.name == header_file.module:
                for sip_file in mod.content:
                    if sip_file.name == header_file.name:
                        for code in list(sip_file.content):
                            if code.status == 'ignored':
                                # Remove any ignored API elements.
                                sip_file.content.remove(code)
                                self._tool.shell.notify(
                                        EventType.CONTAINER_API_DELETE,
                                        (sip_file, code))
                            else:
                                # Mark any non-ignored API elements so that the
                                # user can decide what to do.
                                code.status = 'removed'
                                self._tool.shell.notify(EventType.API_STATUS,
                                        code)

    def _scan_header_file(self, header_path):
        """ Scan a header file and return the header file instance. """

        shell = self._tool.shell

        # Calculate the MD5 signature ignoring any comments.  Note that nested
        # C style comments aren't handled very well.
        m = hashlib.md5()

        src, _, encoding = self._read_header(header_path)

        lnr = 1
        state = 'copy'
        copy = ''
        idx = 0

        for ch in src:
            # Get the previous character.
            if idx > 0:
                prev = src[idx - 1]
            else:
                prev = ''

            idx += 1

            # Line numbers must be accurate.
            if ch == '\n':
                lnr += 1

            # Handle the end of a C style comment.
            if state == 'ccmnt':
                if ch == '/' and prev == '*':
                    state = 'copy'

                continue

            # Handle the end of a C++ style comment.
            if state == 'cppcmnt':
                if ch == '\n':
                    state = 'copy'

                continue

            # We must be in the copy state.

            if ch == '*' and prev == '/':
                # The start of a C style comment.
                state = 'ccmnt'
                continue

            if ch == '/' and prev == '/':
                # The start of a C++ style comment.
                state = 'cppcmnt'
                continue

            # At this point we know the previous character wasn't part of a
            # comment.
            if prev:
                m.update(prev.encode(encoding))

        # Note that we didn't add the last character, but it would normally be
        # a newline.
        md5 = m.hexdigest()

        # See if we already know about the file.
        header_directory = self._header_directory
        header_file_name = os.path.basename(header_path)

        for header_file in header_directory.content:
            if header_file.name == header_file_name:
                new_header_file = False
                break
        else:
            # It's a new file.
            header_file = HeaderFile(name=header_file_name)
            new_header_file = True

        # See if we already know about this version.
        working_version = self._working_version.currentText()

        for header_file_version in header_file.versions:
            if header_file_version.version == working_version:
                # See if the version's contents have changed.
                if header_file_version.md5 != md5:
                    header_file_version.md5 = md5
                    header_file_version.parse = True
                    self._tool.header_file_status(header_file)
                    shell.dirty = True

                break
        else:
            # It's a new version.
            header_file_version = HeaderFileVersion(md5=md5, parse=True,
                    version=working_version)
            header_file.versions.append(header_file_version)

            # Check that the project has versions.
            if len(shell.project.versions) != 0:
                # Find the immediately preceding version if there is one.
                versions_sorted = sorted(header_file.versions,
                        key=lambda v: shell.project.versions.index(v.version))

                prev_md5 = ''
                prev_parse = True
                for hfv in versions_sorted:
                    if hfv.version == working_version:
                        break

                    prev_md5 = hfv.md5
                    prev_parse = hfv.parse

                if prev_md5 == md5:
                    header_file_version.parse = prev_parse
            else:
                # It must be a new file of an unversioned project.
                header_file_version.parse = True

            if new_header_file:
                self._tool.header_file_added(header_file, header_directory,
                        working_version)
            else:
                self._tool.header_file_status(header_file)

            shell.dirty = True

        return header_file

    def _set_module_selector(self, ignored):
        """ Set the module selector for a header file. """

        self._module.clear()

        if not ignored:
            for module in self._tool.shell.project.modules:
                self._module.addItem(module.name, module)

            self._module.model().sort(0)

            # See if we have just un-ignored the file.
            module_name = self._header_file.module

            if module_name == '':
                # If there is a module with the same name of the header
                # directory then assume that's where the file is going.
                for module in self._tool.shell.project.modules:
                    if module.name == self._header_directory.name:
                        module_name = module.name
                        break

            self._module.setCurrentText(module_name)

    def _get_platform_widgets(self):
        """ A generator that returns a 3-tuple of the name and each field
        widget for each supported platform.
        """

        for i in range(self._platforms.count()):
            platform_name = self._platforms.tabText(i)

            page = self._platforms.widget(i)
            inputdirpattern = page.findChild(QLineEdit, 'inputdirpattern')
            parserargs = page.findChild(QLineEdit, 'parserargs')

            yield (platform_name, inputdirpattern, parserargs)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass


@dataclass
class VersionRange:
    """ This class implements a range of versions. """

    # The end version.
    endversion: str = ''

    # The start version.
    startversion: str = ''
